from __future__ import annotations

from dataclasses import dataclass
from enum import auto
from typing import TYPE_CHECKING

from flexohr.harmony.enums import FancyStrEnum

if TYPE_CHECKING:
    pass


class ToneFunction(FancyStrEnum):
    root = auto()
    bass = auto()
    third = auto()
    fifth = auto()
    seventh = auto()
    ninth = auto()
    eleventh = auto()
    thirteenth = auto()
    leading_tone = auto()
    downward_leading_tone = auto()
    altered = auto()
    added_second = auto()
    added_fourth = auto()
    added_sixth = auto()
    second_suspension = auto()
    fourth_suspension = auto()
    pedal = auto()
    undefined = auto()
    # Aliases
    rt = "root"
    bs = "bass"
    _3rd = "third"
    _5th = "fifth"
    _7th = "seventh"
    _9th = "ninth"
    _11th = "eleventh"
    _13th = "thirteenth"
    lt = "leading_tone"
    dlt = "downward_leading_tone"
    alt = "altered"
    a2 = "added_second"
    a4 = "added_fourth"
    a6 = "added_sixth"
    s2 = "second_suspension"
    s4 = "fourth_suspension"
    suspended_second = "second_suspension"
    suspended_fourth = "fourth_suspension"
    ped = "pedal"
    u = "undefined"


class ChordQuality(FancyStrEnum):
    # Triads
    major_triad = auto()
    minor_triad = auto()
    diminished_triad = auto()
    augmented_triad = auto()
    # Sevenths
    dominant_seventh = auto()
    minor_seventh = auto()
    major_seventh = auto()
    minor_major_seventh = auto()
    diminished_seventh = auto()
    half_diminished_seventh = auto()
    augmented_seventh = auto()
    augmented_major_seventh = auto()
    # Augmented sixths
    italian_sixth = auto()
    german_sixth = auto()
    french_sixth = auto()
    # Suspended
    suspended_second = auto()
    suspended_fourth = auto()
    # Power / single / unclassified
    power_chord = auto()
    unison = auto()
    unclassified = auto()
    # Aliases
    maj = "major_triad"
    min = "minor_triad"
    dim = "diminished_triad"
    aug = "augmented_triad"
    Mm7 = "dominant_seventh"
    mm7 = "minor_seventh"
    MM7 = "major_seventh"
    mM7 = "minor_major_seventh"
    o7 = "diminished_seventh"
    hd7 = "half_diminished_seventh"
    aug7 = "augmented_seventh"
    augM7 = "augmented_major_seventh"
    It6 = "italian_sixth"
    Ger6 = "german_sixth"
    Fr6 = "french_sixth"
    sus2 = "suspended_second"
    sus4 = "suspended_fourth"
    pow = "power_chord"


class ChordClass(FancyStrEnum):
    dyad = auto()
    triad = auto()
    seventh = auto()
    ninth = auto()
    eleventh = auto()
    thirteenth = auto()
    augmented_sixth = auto()
    suspended = auto()
    cluster = auto()
    unclassified = auto()
    # Aliases
    _2 = "dyad"
    _3 = "triad"
    _7th = "seventh"
    _9th = "ninth"
    _11th = "eleventh"
    _13th = "thirteenth"
    aug6 = "augmented_sixth"
    sus = "suspended"
    cl = "cluster"


class Inversion(FancyStrEnum):
    root_position = auto()
    first = auto()
    second = auto()
    third = auto()
    fourth = auto()
    # Aliases
    rp = "root_position"
    _1st = "first"
    _2nd = "second"
    _3rd = "third"
    _4th = "fourth"


class CollectionType(FancyStrEnum):
    # Tone fields
    chromatic = auto()
    octatonic_hw = auto()
    octatonic_wh = auto()
    diatonic = auto()
    whole_tone = auto()
    hexatonic = auto()
    pentatonic = auto()
    # Modal / scale types
    major = auto()
    natural_minor = auto()
    harmonic_minor = auto()
    melodic_minor = auto()
    dorian = auto()
    phrygian = auto()
    lydian = auto()
    mixolydian = auto()
    aeolian = auto()
    locrian = auto()
    major_pentatonic = auto()
    minor_pentatonic = auto()
    # Aliases
    chr = "chromatic"
    oct_hw = "octatonic_hw"
    oct_wh = "octatonic_wh"
    dia = "diatonic"
    wt = "whole_tone"
    hex = "hexatonic"
    pent = "pentatonic"
    M = "major"
    m = "natural_minor"
    hm = "harmonic_minor"
    mm = "melodic_minor"
    dor = "dorian"
    phr = "phrygian"
    lyd = "lydian"
    mix = "mixolydian"
    aeo = "aeolian"
    loc = "locrian"
    Mp = "major_pentatonic"
    mp = "minor_pentatonic"


class KeyMode(FancyStrEnum):
    major = auto()
    minor = auto()
    # Aliases
    M = "major"
    m = "minor"


@dataclass(frozen=True)
class ToneModification:
    """Modification to a tone beyond the base chord quality."""

    tone_function: ToneFunction
    alteration: object = None  # SIC when pitchspace available
    added: bool = False
    replacement: object = None  # AbsentComponent when applicable



import re
from functools import cache
from typing import Literal, Optional, Sequence, Tuple, Union, overload


def parse_interval(interval_string: str) -> Tuple[int, int, int]:
    """
    Parse a string representing a spelled interval. Returns a tuple (sign, octave, fifths), where
    sign is +1 or -1 and indicates the sign given in the string (no sign means positive), octave indicates the
    number of full octave steps (in positive or negative direction), and fifths
    indicates the steps taken along the line of fifths (i.e. not actual fifth steps that would add to the octaves).
    """
    if not isinstance(interval_string, str):
        raise TypeError(f"expected string as input, got {type(interval_string).__name__}")
    match = re.match(
        "^(?P<sign>[-+])?("
        "(?P<quality0>P)(?P<generic0>[145])|"  # perfect intervals
        "(?P<quality1>|([Mm]))(?P<generic1>[2367])|"  # imperfect intervals
        "(?P<quality2>(a+|d+))(?P<generic2>[1-7])"  # augmeted/diminished intervals
        ")(?P<octave>(:-?[0-9]+)?)$",
        interval_string,
    )
    if match is None:
        raise ValueError(f"{interval_string!r} is not a valid interval string.")
    # get quality and generic interval (first corresponding group that is not None)
    for i in range(3):
        g = match[f"generic{i}"]
        q = match[f"quality{i}"]
        if g is not None and q is not None:
            generic = int(g)
            quality = q
            break
    else:
        raise RuntimeError(
            f"Could not match generic interval and quality, this is a bug in the regex ("
            f"{[match[f'generic{i}'] for i in range(3)]}, "
            f"{[match[f'quality{i}'] for i in range(3)]}"
            f")"
        )
    # initialise value with generic interval classes
    fifths = (2 * generic - 1) % 7 - 1
    # add modifiers
    if quality in ["P", "M"]:
        pass
    elif quality == "m":
        fifths -= 7
    elif "a" in quality:
        fifths += 7 * len(quality)
    elif "d" in quality:
        if generic in [4, 1, 5]:
            fifths -= 7 * len(quality)
        else:
            fifths -= 7 * (len(quality) + 1)
    else:
        raise RuntimeError(
            f"Initialization from string failed: "
            f"Unexpected interval quality '{quality}'. This is a bug and "
            f"means that either the used regex is bad or the handling code."
        )
    # get octave
    if match["octave"][1:] == "":
        octave = 0
    else:
        octave = int(match["octave"][1:])
    if match["sign"] == "-":
        sign = -1
    else:
        sign = 1
    return sign, fifths, octave


@overload
def split_note_name(note_name: str, count: Literal[False]) -> Tuple[str, str]: ...


@overload
def split_note_name(note_name: str, count: Literal[True]) -> Tuple[int, str]: ...


def split_note_name(note_name: str, count: bool = False) -> Union[Tuple[int, str], Tuple[str, str]]:
    """Splits a note name such as 'Ab' into accidentals and name.

    Args:
        note_name: Note name.
        count: Pass True to get the accidentals as integer rather than as string.
    """
    assert isinstance(note_name, str), f"'{note_name}' is not an accepted note name."
    m = re.match(r"^([A-G]|[a-g])(#*|b*)$", note_name)
    assert m is not None, f"{note_name} is not a valid note name."
    note_name, accidentals = m.group(1), m.group(2)
    if count:
        accidentals = accidentals.count("#") - accidentals.count("b")
    return accidentals, note_name


def split_pitch_name(pitch_name: str) -> Tuple[str, int]:
    """Splits a pitch name such as 'Ab4' into the pitch class name and octave."""
    pitch_name = pitch_name.replace("\u266d", "b")
    pitch_name = pitch_name.replace("\u266f", "#")
    match = re.match(r"(?P<class>[A-Ga-g](?:b*|#*))(?P<octave>-?\d+)", pitch_name)
    if match is None:
        raise ValueError(f"Invalid pitch string: {pitch_name}")
    pitch_class, octave = match["class"], int(match["octave"])
    return pitch_class, octave


def _fifths2str(fifths: int, steps: Sequence[str], inverted: bool = False) -> str:
    """Boilerplate used by fifths2-functions.

    Args:
        fifths: Fifth to be converted to a string.
        steps: List of seven strings.
        inverted: By default, the accidentals are appended. Pass True to prepend them instead.
    """
    fifths += 1
    acc = abs(fifths // 7) * "b" if fifths < 0 else fifths // 7 * "#"
    if inverted:
        return steps[fifths % 7] + acc
    return acc + steps[fifths % 7]


@cache
def fifths2note_name(fifths: int) -> str:
    """Return note name of a stack of fifths such that
       0 = C, -1 = F, -2 = Bb, 1 = G etc.
       Uses: _fifths2str()

    Args:
        fifths: Fifths to be turned into a note name.
    """
    note_names = ["F", "C", "G", "D", "A", "E", "B"]
    name = _fifths2str(fifths, note_names, inverted=True)
    return name


@cache
def steps2note_name(steps: int) -> str:
    """Return note name for a step value such that
       0 = C, 1 = D, 2 = E, etc.

    Args:
        steps: Step [0-6] to be turned into a note name.
    """
    note_names = ["C", "D", "E", "F", "G", "A", "B"]
    return note_names[steps]


def fifths2interval_name(
    fifths: int,
    smallest: bool = False,
    perfect: str = "P",
    major: str = "M",
    minor: str = "m",
    augmented: str = "a",
    diminished: str = "d",
) -> str:
    """Return interval name of a stack of fifths such that 0 = 'P1', -1 = 'P4', -2 = 'm7', 4 = 'M3' etc. If you pass
    ``smallest=True``, intervals of a fifth or greater will be inverted (e.g. 'm6' => '-M3' and 'D5' => '-A4').


    Args:
        fifths: Number of fifths representing the interval
        smallest: Pass True if you want to wrap intervals of a fifths and larger to the downward counterpart.
        perfect: String representing the perfect interval quality, defaults to 'P'.
        major: String representing the major interval quality, defaults to 'M'.
        minor: String representing the minor interval quality, defaults to 'm'.
        augmented: String representing the augmented interval quality, defaults to 'a'.
        diminished: String representing the diminished interval quality, defaults to 'd'.

    Returns:
        Name of the interval as a string.
    """
    fifths_plus_one = fifths + 1  # making 0 = fourth, 1 = unison, 2 = fifth etc.
    int_num = ["4", "1", "5", "2", "6", "3", "7"][fifths_plus_one % 7]
    sharp_wise_quality = augmented
    flat_wise_quality = diminished
    qualities = (
        minor,
        minor,
        minor,
        minor,
        perfect,
        perfect,
        perfect,
        major,
        major,
        major,
        major,
    )
    quality = ""
    if smallest and int(int_num) > 4:
        int_num = str(9 - int(int_num))
        sharp_wise_quality, flat_wise_quality = flat_wise_quality, sharp_wise_quality
        qualities = tuple(reversed(qualities))
        quality = "-"
    if -5 <= fifths <= 5:
        quality += qualities[fifths + 5]
    elif fifths > 5:
        quality += sharp_wise_quality * (fifths_plus_one // 7)
    else:
        quality += flat_wise_quality * ((-fifths + 1) // 7)
    return quality + int_num


def fifths2semitones(fifths: int) -> int:
    """Return the number of semitones in a stack of fifths such that
       C (0) -> 0, G (1) -> 7, D (2) -> 2 etc.

    Args:
        fifths: Fifths to be turned into a number of semitones.
    """
    return fifths * 7 % 12


def fifths2steps(fifths: int) -> int:
    """Return the number of steps in a stack of fifths such that
       C (0) -> 0, D (2) -> 1, E (4) -> 2 etc.

    Args:
        fifths: Fifths to be turned into a number of steps.
    """
    return fifths * 4 % 7


def note_name2fifths(nn):
    """Turn a note name such as `Ab` into a tonal pitch class, such that -1=F, 0=C, 1=G etc.
    Uses: split_note_name()
    """
    name_tpcs = {"C": 0, "D": 2, "E": 4, "F": -1, "G": 1, "A": 3, "B": 5}
    accidentals, note_name = split_note_name(nn, count=True)
    if note_name is None:
        return None
    step_tpc = name_tpcs[note_name.upper()]
    return step_tpc + 7 * accidentals


def note_name2step(note_name) -> Optional[int]:
    """Turn a note name such as `A` into a step, such that C=0, D=1, E=2 etc."""
    name_steps = {"C": 0, "D": 1, "E": 2, "F": 3, "G": 4, "A": 5, "B": 6}
    _, step_name = split_note_name(note_name, count=False)
    if step_name is None:
        return None
    return name_steps[step_name.upper()]


def parse_generic_pitch(pitch_string: str) -> Tuple[int, int]:
    try:
        pitch_class, octave = split_pitch_name(pitch_string)
    except ValueError:
        if pitch_string.startswith("g"):
            pitch_class, octave = split_pitch_name(pitch_string[1:])
        else:
            raise
    step = note_name2step(pitch_class)
    if step is None:
        raise ValueError(f"Invalid pitch string: {pitch_string}")
    return step, octave


def name2semitones(name: str) -> int:
    """Turns an interval or note name into a number of semitones.

    Raises:
        ValueError
    """
    note_name_match = re.match(r"^[A-Ga-g][#b]*$", name)
    interval_name_match = re.match(r"^(P|M|m|a+|d+)(\d+)$", name)
    midi_value_match = re.match(r"^\d+$", name)

    if interval_name_match and not note_name_match and not midi_value_match:
        quality, number = interval_name_match.groups()
        number = int(number)

        semitones_map = {1: 0, 2: 2, 3: 4, 4: 5, 5: 7, 6: 9, 7: 11, 8: 12}

        base_semitones = semitones_map[number]

        if "m" in quality:  # Minor
            base_semitones -= 1
        elif "a" in quality:  # Augmented
            base_semitones += quality.count("a")
        elif "d" in quality:  # Diminished
            base_semitones -= quality.count("d")

        return base_semitones
    elif note_name_match and not interval_name_match and not midi_value_match:
        note_name_tpcs = {"C": 0, "D": 2, "E": 4, "F": 5, "G": 7, "A": 9, "B": 11}
        accidentals, note_name = split_note_name(name, count=True)
        step_tpc = note_name_tpcs[note_name.upper()]

        return (step_tpc + accidentals) % 12
    elif midi_value_match and not note_name_match and not interval_name_match:
        return int(name) % 12
    else:
        raise ValueError(f"'{name}' does not match known patterns for notes or intervals.")


def name2fifths(name: str) -> int:
    """Turns an interval or note name into a number of fifths.

    Raises:
        ValueError
    """
    note_name_match = re.match(r"^[A-Ga-g][#b]*$", name)
    interval_name_match = re.match(r"^(P|M|m|a+|d+)(\d+)$", name)
    int_string_match = re.match(r"^\d+$", name)

    if interval_name_match and not note_name_match and not int_string_match:
        quality, int_num = interval_name_match.group(1), int(interval_name_match.group(2))
        assert int_num > 0, f"'{name}': the interval needs to be non-zero."
        fifths_base = (2 * int_num - 1) % 7 - 1  # base interval class, i.e. P or M version, as fifths
        if quality == "P":
            assert fifths_base in (
                -1,
                0,
                1,
            ), f"'{name}': Only 4ths, 5ths and unisons can be P"
        if quality == "M":
            assert fifths_base in (
                2,
                3,
                4,
                5,
            ), f"'{name}': Only 2nds, 3rds, 6ths, and 7ths can be M"
        elif quality == "m":
            assert fifths_base in (
                2,
                3,
                4,
                5,
            ), f"'{name}': Only 2nds, 3rds, 6ths, and 7ths can be m"
            fifths_base -= 7
        elif "a" in quality:
            fifths_base += 7 * len(quality)
        elif "d" in quality:
            if fifths_base in (-1, 0, 1):
                fifths_base -= 7 * (len(quality))
            else:
                fifths_base -= 7 * (len(quality) + 1)
        return fifths_base
    elif note_name_match and not interval_name_match and not int_string_match:
        note_name_tpcs = {"C": 0, "D": 2, "E": 4, "F": -1, "G": 1, "A": 3, "B": 5}
        accidentals, note_name = split_note_name(name, count=True)
        step_tpc = note_name_tpcs[note_name.upper()]

        return step_tpc + 7 * accidentals
    elif int_string_match and not note_name_match and not interval_name_match:
        return int(name)
    else:
        raise ValueError(f"'{name}' does not match known patterns for notes or intervals.")


def name2steps(name: str) -> int:
    if name.isdigit():
        return int(name) % 7

    note_name_match = re.match(r"^[A-Ga-g][#b]*$", name)
    interval_name_match = re.match(r"^(P|M|m|a+|d+)(\d+)$", name)

    note_name_tpcs = {"C": 0, "D": 1, "E": 2, "F": 3, "G": 4, "A": 5, "B": 6}
    interval_steps = {1: 0, 2: 2, 3: 4, 4: 5, 5: 7, 6: 9, 7: 11}

    if note_name_match:
        note_name = re.sub(r"[#b]+", "", name)
        step_tpc = note_name_tpcs[note_name.upper()]
        return step_tpc
    elif interval_name_match:
        quality, number = interval_name_match.groups()
        number = int(number)

        if number not in interval_steps:
            raise ValueError(f"Interval number '{number}' is not valid.")

        steps = interval_steps[number] % 7

        return steps
    else:
        raise ValueError(f"'{name}' does not match known patterns for notes, intervals, or direct step values.")


def semitone_difference(s1, octave1, s2, octave2) -> int:
    """
    Returns the semitone difference between 2 objects

    Args:
        s1: Semitones of the first object.
        octave1: Octave of the first object.
        s2: Semitones of the second object.
        octave2: Octave of the second object.
    """
    sd = s1 - s2 + (octave1 - octave2) * 12
    return sd


def step_difference(s1, octave1, s2, octave2) -> int:
    """
    Returns the step difference between 2 objects

    Args:
        s1: Steps of the first object.
        octave1: Octave of the first object.
        s2: Steps of the second object.
        octave2: Octave of the second object.
    """
    sd = s1 - s2 + (octave1 - octave2) * 7
    return sd


def interval_tuple_from_sd(sd, factor) -> Tuple[int, int]:
    """
    Returns an interval tuple from steps or semitones in the subtraction operation

    Args:
        sd: Steps or semitones of the interval.
        factor: Factor to be used in the modulo operation (7 for steps, 12 for semitones).
    """
    return sd % factor, sd // factor


def general_tuple(s1, octave1, s2, octave2, factor, sign) -> Tuple[int, int]:
    """
    Returns a pitch tuple from steps or semitones

    Args:
        s1: Steps or semitones of the first object.
        octave1: Octave of the first object.
        s2: Steps or semitones of the second object.
        octave2: Octave of the second object.
        factor: Factor to be used in the modulo operation.
        sign: Sign to be used in the octave calculation (addition or subtraction).
    """
    return (s1 + sign * s2) % factor, octave1 + sign * octave2 + (s1 + sign * s2) // factor


def specific_tuple(fifths1, s1, octave1, fifths2, s2, octave2, sign) -> Tuple[int, int]:
    """
    Returns a tuple from specific types

    Args:
        fifths1: Fifths of the first object.
        s1: Steps or semitones of the first object.
        octave1: Octave of the first object.
        fifths2: Fifths of the second object.
        s2: Steps or semitones of the second object.
        octave2: Octave of the second object.
        sign: Sign to be used in the octave calculation (addition or subtraction).
    """
    return fifths1 + sign * fifths2, octave1 + sign * octave2 + (s1 + sign * s2) // 12



"""Storage experiment harness — benchmarks both backends on common queries.

Generates OHR datasets via the sampler and measures insert time, query time,
round-trip fidelity, and memory for both FlatComponentTable and JsonDictBackend.
"""

from __future__ import annotations

import logging
import time
from dataclasses import dataclass
from typing import Any

from flexohr.core.ohr import OHR
from flexohr.core.sampler import OHRSampler, SamplerConfig
from flexohr.storage.flat_table import FlatComponentTable
from flexohr.storage.json_backend import JsonDictBackend

logger = logging.getLogger(__name__)


@dataclass
class BenchmarkResult:
    """Timing and memory results for a single benchmark run."""

    backend: str
    dataset_size: int
    insert_ms: float
    ohr_query_ms: float
    component_query_ms: float
    roundtrip_ms: float
    roundtrip_ok: bool
    memory_bytes: int
    component_count: int
    ohr_node_count: int

    def summary_dict(self) -> dict[str, Any]:
        return {
            "backend": self.backend,
            "n_ohrs": self.dataset_size,
            "n_components": self.component_count,
            "n_ohr_nodes": self.ohr_node_count,
            "insert_ms": round(self.insert_ms, 2),
            "ohr_query_ms": round(self.ohr_query_ms, 2),
            "comp_query_ms": round(self.component_query_ms, 2),
            "roundtrip_ms": round(self.roundtrip_ms, 2),
            "roundtrip_ok": self.roundtrip_ok,
            "memory_bytes": self.memory_bytes,
        }


class StorageExperiment:
    """Run storage benchmarks on both backends with sampler-generated OHRs."""

    def __init__(self, seed: int = 42) -> None:
        self.seed = seed

    def generate_dataset(self, n: int, config: SamplerConfig | None = None) -> list[OHR]:
        """Generate *n* OHRs using the sampler."""
        sampler = OHRSampler(config=config, seed=self.seed)
        return sampler.sample_batch(n)

    def benchmark_flat_table(self, ohrs: list[OHR]) -> BenchmarkResult:
        """Benchmark the flat component-per-row table backend."""
        # Insert
        t0 = time.perf_counter()
        table = FlatComponentTable()
        table.add_ohrs(ohrs)
        insert_ms = (time.perf_counter() - t0) * 1000

        # OHR-level query: find OHRs with a specific property value
        # Use a common chord quality as the target
        t0 = time.perf_counter()
        table.query_ohrs_by_property("chord_quality", "major_triad")
        ohr_query_ms = (time.perf_counter() - t0) * 1000

        # Component-level query: find all roots
        t0 = time.perf_counter()
        table.query_components_by_property("tone_function", "root")
        component_query_ms = (time.perf_counter() - t0) * 1000

        # Round-trip: reconstruct all top-level OHRs and compare dicts
        t0 = time.perf_counter()
        reconstructed = table.reconstruct_all_top_level()
        roundtrip_ms = (time.perf_counter() - t0) * 1000

        # Check round-trip by comparing dict representations
        roundtrip_ok = self._check_roundtrip(ohrs, reconstructed)

        return BenchmarkResult(
            backend="flat_table",
            dataset_size=len(ohrs),
            insert_ms=insert_ms,
            ohr_query_ms=ohr_query_ms,
            component_query_ms=component_query_ms,
            roundtrip_ms=roundtrip_ms,
            roundtrip_ok=roundtrip_ok,
            memory_bytes=table.memory_bytes,
            component_count=table.count_components(),
            ohr_node_count=table.count_ohrs(),
        )

    def benchmark_json_backend(self, ohrs: list[OHR]) -> BenchmarkResult:
        """Benchmark the JSON dict list backend."""
        # Insert
        t0 = time.perf_counter()
        backend = JsonDictBackend()
        backend.add_ohrs(ohrs)
        insert_ms = (time.perf_counter() - t0) * 1000

        # OHR-level query
        t0 = time.perf_counter()
        backend.query_ohrs_by_property("chord_quality", "major_triad")
        ohr_query_ms = (time.perf_counter() - t0) * 1000

        # Component-level query
        t0 = time.perf_counter()
        backend.query_components_by_property("tone_function", "root")
        component_query_ms = (time.perf_counter() - t0) * 1000

        # Round-trip
        t0 = time.perf_counter()
        reconstructed = backend.reconstruct_all()
        roundtrip_ms = (time.perf_counter() - t0) * 1000

        roundtrip_ok = self._check_roundtrip(ohrs, reconstructed)

        return BenchmarkResult(
            backend="json_dict",
            dataset_size=len(ohrs),
            insert_ms=insert_ms,
            ohr_query_ms=ohr_query_ms,
            component_query_ms=component_query_ms,
            roundtrip_ms=roundtrip_ms,
            roundtrip_ok=roundtrip_ok,
            memory_bytes=backend.memory_bytes,
            component_count=backend.count_components(),
            ohr_node_count=backend.count_ohrs(),
        )

    def run(
        self,
        sizes: list[int] | None = None,
        config: SamplerConfig | None = None,
    ) -> list[BenchmarkResult]:
        """Run full benchmark suite on both backends at multiple dataset sizes."""
        if sizes is None:
            sizes = [100, 1000]

        results: list[BenchmarkResult] = []
        for n in sizes:
            logger.info("Benchmarking with %d OHRs", n)
            ohrs = self.generate_dataset(n, config)

            results.append(self.benchmark_flat_table(ohrs))
            results.append(self.benchmark_json_backend(ohrs))

        return results

    def results_to_dataframe(self, results: list[BenchmarkResult]) -> Any:
        """Convert results to a pandas DataFrame for display."""
        import pandas as pd

        return pd.DataFrame([r.summary_dict() for r in results])

    # ── Internal ──

    def _check_roundtrip(self, originals: list[OHR], reconstructed: list[OHR]) -> bool:
        """Check round-trip fidelity by comparing dict representations."""
        if len(originals) != len(reconstructed):
            return False
        for orig, recon in zip(originals, reconstructed):
            if orig.to_dict() != recon.to_dict():
                return False
        return True



from __future__ import annotations

from typing import Any, Tuple, Union

from .abstract import (
    FifthsScalar,
    Interval,
    IntervalClass,
    IntType,
    Octaves,
    SemitonesScalar,
    StepsScalar,
)
from .errors import InvalidInitValue, StepUnknown
from .utils import (
    fifths2interval_name,
    fifths2semitones,
    fifths2steps,
    general_tuple,
    interval_tuple_from_sd,
    parse_interval,
    semitone_difference,
    specific_tuple,
    step_difference,
)


class EnharmonicIntervalClass(SemitonesScalar, IntervalClass):
    """Enharmonic Interval Class (equivalent to the IC of pitch-class set theory).
    An EIC value is always within [0, 11] and can be interpreted as the distance between two piano keys when
    projected into the same octave.

    Initialized from an integer within [0, 11]. Other values will be converted. For instance:

        * integers outside this interval are transformed with modulo 12, e.g. EIC(-3) => EIC(9)
        * floats are converted to an integer (i.e., rounded)

    Integers represented as string are accepted, too.
    """

    def __str__(self):
        return f"EIC({self.semitones})"

    def __repr__(self):
        return f"EIC({self.semitones})"

    def _repr_html_(self):
        from flexohr.core.html_colors import PITCHSPACE_COLORS

        c = PITCHSPACE_COLORS["EIC"]
        return f'<span style="font-family:monospace;color:{c}"><b>EIC</b>(<i>{self.semitones}</i>)</span>'

    def __neg__(self):
        return EnharmonicIntervalClass(-self.semitones)

    def __add__(self, other: Union[Interval, IntervalClass, int, IntType]) -> Union[IntervalClass]:
        if isinstance(
            other,
            EnharmonicInterval | EnharmonicIntervalClass | SpecificInterval | SpecificIntervalClass,
        ):
            return EnharmonicIntervalClass((self.semitones + other.semitones) % 12)
        elif isinstance(other, GenericInterval | GenericIntervalClass):
            raise StepUnknown(f"{type(self)} + {type(other)}")

        return super().__add__(other)

    def __sub__(self, other: Union[Interval, IntervalClass, int, IntType]) -> Union[IntervalClass]:
        if isinstance(
            other,
            EnharmonicInterval | EnharmonicIntervalClass | SpecificInterval | SpecificIntervalClass,
        ):
            return EnharmonicIntervalClass((self.semitones - other.semitones) % 12)
        elif isinstance(other, GenericInterval | GenericIntervalClass):
            raise StepUnknown(f"{type(self)} - {type(other)}")

        return super().__sub__(other)

    def to_point(self) -> "EnharmonicPitchClass":  # noqa: F821
        from .pitch import EnharmonicPitchClass

        return EnharmonicPitchClass(self.semitones)


class SpecificIntervalClass(FifthsScalar, IntervalClass):
    """Specific Interval Class, also called tonal interval class, or simply 'interval name'. An SIC is
    numerically represented as a stack of perfect fifths, e.g. SIC("P4")._value == -1

    Initialized from an interval name (e.g. "M3" or "m6") or from an integer representing a stack of
    perfect fifths. Integers represented as string are accepted, too. Floats are
    converted to an integer (i.e., rounded).
    """

    @property
    def name(self):
        return fifths2interval_name(self.fifths)

    def __neg__(self):
        return SpecificIntervalClass(-self.fifths)

    def __str__(self):
        return f"SIC('{self.name}')"

    def __repr__(self):
        return self.name

    def _repr_html_(self):
        from flexohr.core.html_colors import PITCHSPACE_COLORS

        c = PITCHSPACE_COLORS["SIC"]
        return f'<span style="font-family:monospace;color:{c}"><b>SIC</b>(<i>{self.name}</i>)</span>'

    def __add__(self, other: Union[Interval, IntervalClass, int, IntType]) -> Union[IntervalClass]:
        if isinstance(other, EnharmonicInterval | EnharmonicIntervalClass):
            return EnharmonicIntervalClass((self.semitones + other.semitones) % 12)
        elif isinstance(other, SpecificInterval | SpecificIntervalClass):
            return SpecificIntervalClass(self.fifths + other.fifths)
        elif isinstance(other, GenericInterval | GenericIntervalClass):
            return GenericIntervalClass((self.steps + other.steps) % 7)

        return super().__add__(other)

    def __sub__(self, other: Union[Interval, IntervalClass, int, IntType]) -> Union[IntervalClass]:
        if isinstance(other, EnharmonicInterval | EnharmonicIntervalClass):
            return EnharmonicIntervalClass((self.semitones - other.semitones) % 12)
        elif isinstance(other, SpecificInterval | SpecificIntervalClass):
            return SpecificIntervalClass(self.fifths - other.fifths)
        elif isinstance(other, GenericInterval | GenericIntervalClass):
            return GenericIntervalClass((self.steps - other.steps) % 7)

        return super().__sub__(other)

    def to_point(self) -> "SpecificPitchClass":  # noqa: F821
        from .pitch import SpecificPitchClass

        return SpecificPitchClass(self.fifths)


class GenericIntervalClass(StepsScalar, IntervalClass):
    """Generic Interval Class, also called 'interval number'. A GIC value is always within [0, 6] and can be
    interpreted as the distance between white piano keys projected into the same octave.

    Initialized from an integer within [0, 6]. Other values will be converted. For instance:

        * integers outside this interval are transformed with modulo 7, e.g. GIC(-3) => GIC(4)
        * interval names are mapped to their step, e.g. GIC("P5") => GIC(4)
        * floats are converted to an integer (i.e., rounded)

    Integers represented as string are accepted, too.
    """

    def __str__(self):
        return f"GIC({self.steps})"

    def __repr__(self):
        return f"GIC({self.steps})"

    def _repr_html_(self):
        from flexohr.core.html_colors import PITCHSPACE_COLORS

        c = PITCHSPACE_COLORS["GIC"]
        return f'<span style="font-family:monospace;color:{c}"><b>GIC</b>(<i>{self.steps}</i>)</span>'

    def __neg__(self):
        return GenericIntervalClass(-self.steps)

    def __add__(self, other: Union[Interval, IntervalClass, int, IntType]) -> Union[IntervalClass]:
        if isinstance(other, EnharmonicInterval | EnharmonicIntervalClass):
            raise StepUnknown(f"{type(self)} + {type(other)}")
        elif isinstance(
            other,
            GenericInterval | GenericIntervalClass | SpecificInterval | SpecificIntervalClass,
        ):
            return GenericIntervalClass((self.steps + other.steps) % 7)

        return super().__add__(other)

    def __sub__(self, other: Union[Interval, IntervalClass, int, IntType]) -> Union[IntervalClass]:
        if isinstance(other, EnharmonicInterval | EnharmonicIntervalClass):
            raise StepUnknown(f"{type(self)} - {type(other)}")
        elif isinstance(
            other,
            GenericInterval | GenericIntervalClass | SpecificInterval | SpecificIntervalClass,
        ):
            return GenericIntervalClass((self.steps - other.steps) % 7)

        return super().__sub__(other)


class EnharmonicInterval(Interval[EnharmonicIntervalClass, Octaves]):
    """An Enharmonic Interval is an :class:`EnharmonicIntervalClass` paired with an octave number and
    equivalent to a semitone distance (i.e., step and quality are not known).

    Initialized from a pair of two values (EIC, Octave) or from an integer representing a semitone
    distance. When initialized from a pair, the first value can be any value that can be used for
    initializing an EIC. Strings representing interval names (e.g., "P5:1") are also accepted.
    Integers represented as string are accepted, too.
    """

    def __str__(self):
        return f"EI({self.name})"

    def __repr__(self):
        return f"EI({self.name})"

    def _repr_html_(self):
        from flexohr.core.html_colors import PITCHSPACE_COLORS

        c = PITCHSPACE_COLORS["EI"]
        return f'<span style="font-family:monospace;color:{c}"><b>EI</b>(<i>{self.name}</i>)</span>'

    def __neg__(self):
        return EnharmonicInterval((-self.interval_class.semitones, -self.octaves))

    @staticmethod
    def convert_init_value(
        value: Union[
            Tuple[EnharmonicIntervalClass | int, Octaves | int],
            str,
            int,
            "SpecificInterval",
        ],
    ) -> Tuple[EnharmonicIntervalClass, Octaves]:
        """Used by the constructor to convert the initial value into a pair of IntType."""
        if isinstance(value, tuple):
            interval_class, octave = value
            while interval_class < 0:
                interval_class += 12
                octave -= 1
            while interval_class >= 12:
                interval_class -= 12
                octave += 1
        elif (isinstance(value, str) and value.isdigit()) or isinstance(value, int):
            value_int = int(value)
            interval_class = value_int
            octave = value_int // 12
        elif isinstance(value, str):
            sign, fifths, octave = parse_interval(value)
            interval_class = fifths2semitones(fifths)
        else:
            raise InvalidInitValue(
                f"An EnharmonicInterval cannot be expressed as a {type(value)}. Pass an interval name as a string "
                f"or an (IntervalClass, Octave) tuple."
            )
        return EnharmonicIntervalClass(interval_class), Octaves(octave)

    def midi(self) -> int:
        if self.semitones in [0, 127]:
            return self.semitones
        raise ValueError(f"Cannot convert {self} to MIDI number.")

    @property
    def semitones(self) -> int:
        return int(self.octaves) * 12 + self.interval_class.semitones

    @property
    def name(self) -> str:
        return str(self.semitones)

    def __add__(
        self,
        other: Union[
            Interval,
            IntervalClass,
            int,
            IntType,
            Tuple[EnharmonicIntervalClass | int, Octaves | int],
        ],
    ) -> (
        EnharmonicInterval
        | EnharmonicIntervalClass
        | Interval
        | IntervalClass
        | tuple[EnharmonicIntervalClass, Octaves]
        | tuple[IntervalClass | Any, Octaves | Any]
    ):
        if isinstance(other, EnharmonicInterval | SpecificInterval):
            return EnharmonicInterval(
                general_tuple(
                    self.interval_class.semitones,
                    self.octaves,
                    other.interval_class.semitones,
                    other.octaves,
                    12,
                    1,
                )
            )
        elif isinstance(other, EnharmonicIntervalClass | SpecificIntervalClass):
            return EnharmonicIntervalClass((self.interval_class.semitones + other.semitones) % 12)
        elif isinstance(other, GenericIntervalClass | GenericInterval):
            raise StepUnknown(f"{type(self)} + {type(other)}")

        elif Tuple == type(other):
            return self + EnharmonicInterval(other)
        return super().__add__(other)

    def __sub__(
        self,
        other: Union[
            Interval,
            IntervalClass,
            IntType,
            int,
            Tuple[EnharmonicIntervalClass | int, Octaves | int],
        ],
    ) -> (
        EnharmonicInterval
        | EnharmonicIntervalClass
        | int
        | IntType
        | IntervalClass
        | Interval
        | tuple[EnharmonicIntervalClass, Octaves]
        | tuple[IntervalClass | Any, Octaves | Any]
    ):
        if isinstance(other, EnharmonicInterval | SpecificInterval):
            sd = semitone_difference(
                self.interval_class.semitones,
                self.octaves,
                other.interval_class.semitones,
                other.octaves,
            )
            return EnharmonicInterval(interval_tuple_from_sd(sd, 12))
        elif isinstance(other, EnharmonicIntervalClass | SpecificIntervalClass):
            return EnharmonicIntervalClass((self.interval_class.semitones - other.semitones) % 12)
        elif isinstance(other, GenericInterval | GenericIntervalClass):
            raise StepUnknown(f"{type(self)} - {type(other)}")

        elif Tuple == type(other):
            return self - EnharmonicInterval(other)
        return super().__sub__(other)

    def __eq__(self, other):
        if isinstance(other, EnharmonicInterval):
            return self.interval_class.semitones == other.interval_class.semitones and self.octaves == other.octaves
        return False

    def to_point(self) -> "EnharmonicPitchClass":  # noqa: F821
        from .pitch import EnharmonicPitchClass

        return EnharmonicPitchClass(self.semitones)


class SpecificInterval(Interval[SpecificIntervalClass, Octaves]):
    """A Specific Interval, also called "spelled interval", is a :class:`SpecificIntervalClass`
    paired with an octave number. It is the most complete interval representation because number,
    quality, and octave are known.

    Initialized from a pair of two values (SIC, Octave) or from an interval name such as "M3:1".
    When initialized from a pair, the first value can be any value that can be used for
    initializing an SIC.
    """

    @staticmethod
    def convert_init_value(
        value: Union[Tuple[SpecificIntervalClass | int, Octaves | int], str, int],
    ) -> Tuple[SpecificIntervalClass, Octaves]:
        """Used by the constructor to convert the initial value into a pair of IntType."""
        if isinstance(value, tuple):
            interval_class, octave = value
        elif isinstance(value, str):
            sign, interval_class, octave = parse_interval(value)
        elif isinstance(value, int):
            interval_class = value
            octave = 0
        else:
            raise InvalidInitValue(
                f"A SpecificInterval cannot be expressed as a {type(value)}. Pass an interval name as a string "
                f"(e.g., 'a4:1' for an augmented fourth plus an octave) "
                f"or an (IntervalClass, Octave) tuple."
            )
        return SpecificIntervalClass(interval_class), Octaves(octave)

    @property
    def fifths(self) -> int:
        return self.interval_class.fifths

    @property
    def midi(self) -> int:
        """Alias of the semitones property, representing the MIDI number of the pitch."""
        if self.semitones in [0, 127]:
            return self.semitones
        raise ValueError(f"Cannot convert {self} to MIDI number.")

    @property
    def name(self) -> str:
        return f"{self.interval_class.name}:{int(self.octaves)}"

    @property
    def semitones(self) -> int:
        """For an SP, the value of the semitones property correspond to the MIDI number."""
        return int(self.octaves) * 12 + self.interval_class.semitones

    @property
    def steps(self) -> int:
        return int(self.octaves) * 7 + self.interval_class.steps

    def __repr__(self):
        return f"SI({self.name})"

    def __str__(self):
        return f"SI({self.name})"

    def _repr_html_(self):
        from flexohr.core.html_colors import PITCHSPACE_COLORS

        c = PITCHSPACE_COLORS["SI"]
        return f'<span style="font-family:monospace;color:{c}"><b>SI</b>(<i>{self.name}</i>)</span>'

    def __neg__(self):
        return SpecificInterval((-self.fifths, -self.octaves - 1))

    def __add__(
        self,
        other: Union[
            Interval,
            IntervalClass,
            int,
            IntType,
            Tuple[SpecificIntervalClass | int, Octaves | int],
        ],
    ) -> Interval | IntervalClass | tuple[SpecificIntervalClass, Octaves] | tuple[IntervalClass | Any, Octaves | Any]:
        if isinstance(other, EnharmonicInterval):
            return EnharmonicInterval(
                general_tuple(
                    self.interval_class.semitones,
                    self.octaves,
                    other.interval_class.semitones,
                    other.octaves,
                    12,
                    1,
                )
            )
        elif isinstance(other, SpecificInterval):
            return SpecificInterval(
                specific_tuple(
                    self.fifths,
                    self.interval_class.semitones,
                    self.octaves,
                    other.fifths,
                    other.interval_class.semitones,
                    other.octaves,
                    1,
                )
            )
        elif isinstance(other, GenericInterval):
            return GenericInterval(
                general_tuple(
                    self.interval_class.steps,
                    self.octaves,
                    other.interval_class.steps,
                    other.octaves,
                    7,
                    1,
                )
            )
        elif isinstance(other, EnharmonicIntervalClass):
            return EnharmonicIntervalClass((self.interval_class.semitones + other.semitones) % 12)
        elif isinstance(other, SpecificIntervalClass):
            return SpecificIntervalClass(self.fifths + other.fifths)
        elif isinstance(other, GenericIntervalClass):
            return GenericIntervalClass((self.interval_class.steps + other.steps) % 7)

        elif Tuple == type(other):
            return self + SpecificInterval(other)
        return super().__add__(other)

    def __sub__(
        self,
        other: Union[
            Interval,
            IntervalClass,
            int,
            IntType,
            Tuple[SpecificIntervalClass | int, Octaves | int],
        ],
    ) -> Interval | IntervalClass | tuple[SpecificIntervalClass, Octaves] | tuple[IntervalClass | Any, Octaves | Any]:
        if isinstance(other, EnharmonicInterval):
            sd = semitone_difference(
                self.interval_class.semitones,
                self.octaves,
                other.interval_class.semitones,
                other.octaves,
            )
            return EnharmonicInterval(interval_tuple_from_sd(sd, 12))
        elif isinstance(other, SpecificInterval):
            return SpecificInterval(
                specific_tuple(
                    self.fifths,
                    self.interval_class.semitones,
                    self.octaves,
                    other.fifths,
                    other.interval_class.semitones,
                    other.octaves,
                    -1,
                )
            )
        elif isinstance(other, GenericInterval):
            return GenericInterval(
                general_tuple(
                    self.interval_class.steps,
                    self.octaves,
                    other.interval_class.steps,
                    other.octaves,
                    7,
                    -1,
                )
            )
        elif isinstance(other, EnharmonicIntervalClass):
            return EnharmonicIntervalClass((self.interval_class.semitones - other.semitones) % 12)
        elif isinstance(other, SpecificIntervalClass):
            return SpecificIntervalClass(self.fifths - other.fifths)
        elif isinstance(other, GenericIntervalClass):
            return GenericIntervalClass(self.interval_class.steps - other.steps)

        elif Tuple == type(other):
            return self - SpecificInterval(other)
        return super().__sub__(other)

    def __eq__(self, other):
        if isinstance(other, SpecificInterval):
            return self.fifths == other.fifths and self.octaves == other.octaves
        return False

    def to_point(self) -> "SpecificPitchClass":  # noqa: F821
        from .pitch import SpecificPitchClass

        return SpecificPitchClass(self.fifths)


class GenericInterval(Interval[GenericIntervalClass, Octaves]):
    """A Generic Interval is a :class:`GenericIntervalClass` paired with an octave number.
    It represents an interval where the number and the octave are known, but not the quality.
    E.g., think of a generic third as "any third", including minor, major, diminished, etc.

    Initialized from a pair of two values (GIC, Octave) or from an interval name such as "P5:1".
    When initialized from a pair, the first value can be any value that can be used for
    initializing a GIC. Integers are also accepted where the number is the interval step count,
    e.g., GenericInterval(5) for a 5th spanning 0 octaves.
    """

    @staticmethod
    def convert_init_value(
        value: Union[
            Tuple[GenericIntervalClass | int, Octaves | int],
            str,
            int,
            "SpecificInterval",
        ],
    ) -> Tuple[GenericIntervalClass, Octaves]:
        """Used by the constructor to convert the initial value into a pair of IntType."""
        if isinstance(value, tuple):
            interval_class, octave = value
            while interval_class < 0:
                interval_class += 7
                octave -= 1
            while interval_class >= 7:
                interval_class -= 7
                octave += 1
        elif isinstance(value, str):
            sign, fifths, octave = parse_interval(value)
            interval_class = fifths2steps(fifths)
        elif isinstance(value, int):
            interval_class = value % 7
            octave = value // 7
        else:
            raise InvalidInitValue(
                f"A GenericInterval cannot be expressed as a {type(value)}. Pass an interval name as a string "
                f"(e.g., 'P5:1' for a fifth plus an octave) "
                f"or an (GenericInterval, Octave) tuple."
            )
        return GenericIntervalClass(interval_class), Octaves(octave)

    @property
    def name(self) -> str:
        return f"{self.interval_class.name}:{int(self.octaves)}"

    @property
    def steps(self) -> int:
        return int(self.octaves) * 7 + self.interval_class.steps

    def __repr__(self):
        return f"GI({self.name})"

    def __str__(self):
        return f"GI({self.name})"

    def _repr_html_(self):
        from flexohr.core.html_colors import PITCHSPACE_COLORS

        c = PITCHSPACE_COLORS["GI"]
        return f'<span style="font-family:monospace;color:{c}"><b>GI</b>(<i>{self.name}</i>)</span>'

    def __neg__(self):
        return GenericInterval((-self.interval_class.steps, -self.octaves))

    def __add__(
        self,
        other: Union[
            Interval,
            IntervalClass,
            int,
            IntType,
            Tuple[GenericIntervalClass | int, Octaves | int],
        ],
    ) -> (
        GenericInterval
        | GenericIntervalClass
        | Interval
        | IntervalClass
        | tuple[GenericIntervalClass, Octaves]
        | tuple[IntervalClass | Any, Octaves | Any]
    ):
        if isinstance(other, GenericInterval | SpecificInterval):
            return GenericInterval(
                general_tuple(
                    self.interval_class.steps,
                    self.octaves,
                    other.interval_class.steps,
                    other.octaves,
                    7,
                    1,
                )
            )
        elif isinstance(other, GenericIntervalClass | SpecificIntervalClass):
            return GenericIntervalClass((self.interval_class.steps + other.steps) % 7)
        elif isinstance(other, EnharmonicInterval | EnharmonicIntervalClass):
            raise StepUnknown(f"{type(self)} + {type(other)}")

        elif Tuple == type(other):
            return self + GenericInterval(other)
        return super().__add__(other)

    def __sub__(
        self,
        other: Union[
            Interval,
            IntervalClass,
            int,
            IntType,
            Tuple[GenericIntervalClass | int, Octaves | int],
        ],
    ) -> (
        GenericInterval
        | GenericIntervalClass
        | Interval
        | IntervalClass
        | tuple[GenericIntervalClass, Octaves]
        | tuple[IntervalClass | Any, Octaves | Any]
    ):
        if isinstance(other, GenericInterval | SpecificInterval):
            sd = step_difference(
                self.interval_class.steps,
                self.octaves,
                other.interval_class.steps,
                other.octaves,
            )
            return GenericInterval(interval_tuple_from_sd(sd, 7))

        elif isinstance(other, GenericIntervalClass | SpecificIntervalClass):
            return GenericIntervalClass((self.interval_class.steps - other.steps) % 7)
        elif isinstance(other, EnharmonicInterval | EnharmonicIntervalClass):
            raise StepUnknown(f"{type(self)} - {type(other)}")

        elif Tuple == type(other):
            return self - GenericInterval(other)
        return super().__sub__(other)

    def __eq__(self, other):
        if isinstance(other, GenericInterval):
            return self.interval_class.steps == other.interval_class.steps and self.octaves == other.octaves
        return False


EIC = EnharmonicIntervalClass
SIC = SpecificIntervalClass
GIC = GenericIntervalClass
EI = EnharmonicInterval
SI = SpecificInterval
GI = GenericInterval



"""FlexOHR — Flexible and Extensible Object for Harmony Representation."""

from flexohr.core.component import (
    UNDEFINED,
    AbsentComponent,
    Component,
    UndefinedComponent,
)
from flexohr.core.ohr import OHR, Where, WhereSet
from flexohr.core.protocol import FlexohrProtocol
from flexohr.core.sampler import OHRSampler, SamplerConfig
from flexohr.core.validators import (
    BodyNotEmpty,
    HasReference,
    InversionBassConsistency,
    MaxNestingDepth,
    NoContradictingToneFunctions,
    ValidationResult,
    Validator,
)
from flexohr.harmony.enums import (
    EMPTY_PROPS,
    CodecRegistry,
    ComponentStatus,
    FancyStrEnum,
    FrozenProperties,
    KeyValue,
    Source,
    props,
)

C = Component

__all__ = [
    # Core
    "Component",
    "C",
    "AbsentComponent",
    "UndefinedComponent",
    "UNDEFINED",
    "OHR",
    "Where",
    "WhereSet",
    "FlexohrProtocol",
    # Sampler
    "OHRSampler",
    "SamplerConfig",
    # Validators
    "Validator",
    "ValidationResult",
    "HasReference",
    "BodyNotEmpty",
    "InversionBassConsistency",
    "MaxNestingDepth",
    "NoContradictingToneFunctions",
    # Types
    "FancyStrEnum",
    "CodecRegistry",
    "ComponentStatus",
    "Source",
    "FrozenProperties",
    "KeyValue",
    "EMPTY_PROPS",
    "props",
]



from typing import Optional

from flexohr.errors import (
    FlexohrError,
    FlexohrTypeError,
    FlexohrValueError,
    register_paradigm_errors,
)


class CustomException(FlexohrError):
    """Exception that have a default message that can be extended. Default messages
    should be defined without trailing punctuation.
    """

    default_message = "Exception"

    def __init__(self, message: Optional[str] = None):
        if message:
            msg = f"{self.default_message}: {message}"
        else:
            msg = self.default_message + "."
        self.message = msg
        super().__init__(msg)


# region TypeErrors


class CustomTypeError(CustomException, FlexohrTypeError):
    """TypeErrors that have a default message that can be extended. Default messages
    should be defined without trailing punctuation.

    Examples:

        >>> raise CustomTypeError
        TypeError.
        >>> raise CustomTypeError("This is a custom message.")
        TypeError: This is a custom message.

    """

    default_message = "TypeError"


class OperationUndefined(CustomTypeError):
    """Raised when an operation is not defined for the given arguments."""

    default_message = "Operation not defined for the given types"


class PropertyUndefined(CustomTypeError):
    """Raised when a property is not defined for the given type.
    This usually means that an object needs to be converted to another type first.

    Examples:

        >>> EnharmonicPitchClass(1).fifths
        PropertyUndefined: Property not defined for this type: EnharmonicPitchClass
    """

    default_message = "Property not defined for this type"


class StepUnknown(OperationUndefined):
    """Raised when knowing a step (GP or GPC) would be required for an operation."""

    default_message = "Knowing the step would be required for the operation"


# endregion TypeErrors
# region ValueErrors


class CustomValueError(CustomException, FlexohrValueError):
    """ValueErrors that have a default message that can be extended. Default messages
    should be defined without trailing punctuation.
    """

    default_message = "ValueError"


class InvalidInitValue(CustomValueError):
    """Raised when an invalid value is passed to an object's initializer."""

    default_message = "Object cannot be initialized from the given value"

    def __init__(self, message: Optional[str] = None, cls=None, val=None):  # noqa: B042
        """Raised when an invalid value is passed to an object's initializer.

        Args:
            message: Custom message.
            cls: Class that cannot be initialized from the given value.
            val: Invalid value.
        """
        msg = message if message else ""
        if cls or val:
            msg += "\n"
        if cls:
            try:
                msg += f"Class: {cls.__module__}.{cls.__qualname__}"
            except Exception:
                msg += f"Class: {cls}"
            if val:
                msg += ", "
        if val:
            msg += f"Value: {val!r}"
        self.cls = cls
        self.val = val
        self.message = msg
        super().__init__(msg)


class InitValueNotConvertibleToSemitones(InvalidInitValue):
    default_message = "Value cannot be converted to semitones"


class InitValueNotConvertibleToFifths(InvalidInitValue):
    default_message = "Value cannot be converted to fifths"


class InitValueNotConvertibleToStep(InvalidInitValue):
    default_message = "Value cannot be converted to a step A-G"


# endregion ValueErrors


# Register pitchspace errors in the FlexOHR error hierarchy
register_paradigm_errors(
    "pitchspace",
    OperationUndefined,
    PropertyUndefined,
    StepUnknown,
    InvalidInitValue,
    InitValueNotConvertibleToSemitones,
    InitValueNotConvertibleToFifths,
    InitValueNotConvertibleToStep,
)



"""Flat component-per-row table storage model.

Every component across all OHRs lives in a single "long" DataFrame.
OHR structure is reconstructed via graph-like views on the table.

Flat columns (tone_function, etc.) enable fast SQL queries.  The full
serialized properties JSON is preserved alongside for lossless round-trip.
"""

from __future__ import annotations

import json
import logging
from dataclasses import dataclass
from typing import Any

import pandas as pd

from flexohr.core.component import AbsentComponent, Component, UndefinedComponent
from flexohr.core.ohr import (
    OHR,
    _component_from_dict,
    _component_to_dict,
    _properties_from_json,
    _properties_to_json,
)
from flexohr.harmony.enums import (
    EMPTY_PROPS,
    get_value_type_name,
)

logger = logging.getLogger(__name__)


@dataclass
class _OHRMeta:
    """Lightweight OHR-level metadata stored alongside the component table."""

    ohr_id: int
    parent_ohr_id: int | None
    body_position: int | None  # position in parent body (-1 = not a body child)
    ordered: bool
    horizontal: bool
    properties_json: str  # serialized OHR-level properties
    reference_ohr_id: int | None


class FlatComponentTable:
    """Flat component-per-row storage for OHR trees.

    Each row is one component.  OHR structure is metadata that allows
    reconstruction of the original tree.  Properties are stored both as
    flat query columns (tone_function) and as a full JSON blob for lossless
    round-trip.
    """

    def __init__(self) -> None:
        self._rows: list[dict[str, Any]] = []
        self._ohr_meta: dict[int, _OHRMeta] = {}
        self._next_component_id = 0
        self._next_ohr_id = 0

    # ── Ingest ──

    def add_ohr(
        self,
        ohr: OHR,
        parent_ohr_id: int | None = None,
        body_position: int | None = None,
    ) -> int:
        """Flatten *ohr* into the table, returning its ohr_id."""
        ohr_id = self._next_ohr_id
        self._next_ohr_id += 1

        # Handle reference_ohr recursively (not a body child, so body_position=None)
        ref_ohr_id: int | None = None
        if ohr.reference_ohr is not None:
            ref_ohr_id = self.add_ohr(ohr.reference_ohr, parent_ohr_id=ohr_id, body_position=None)

        # Store OHR-level metadata
        ohr_props_json = json.dumps(_properties_to_json(ohr.properties) if ohr.properties else [])
        self._ohr_meta[ohr_id] = _OHRMeta(
            ohr_id=ohr_id,
            parent_ohr_id=parent_ohr_id,
            body_position=body_position,
            ordered=ohr.ordered,
            horizontal=ohr.horizontal,
            properties_json=ohr_props_json,
            reference_ohr_id=ref_ohr_id,
        )

        # Reference component (position=-1 sentinel)
        if ohr.reference_component is not None:
            self._add_component_row(ohr.reference_component, ohr_id, position=-1, is_reference=True)

        # Body items — both Components and nested OHRs get a position
        for pos, item in enumerate(ohr.body):
            if isinstance(item, OHR):
                self.add_ohr(item, parent_ohr_id=ohr_id, body_position=pos)
            else:
                self._add_component_row(item, ohr_id, position=pos, is_reference=False)

        return ohr_id

    def add_ohrs(self, ohrs: list[OHR]) -> list[int]:
        """Add multiple top-level OHRs, returning their ohr_ids."""
        return [self.add_ohr(ohr) for ohr in ohrs]

    def _add_component_row(
        self,
        comp: Component,
        ohr_id: int,
        position: int,
        is_reference: bool,
    ) -> None:
        cid = self._next_component_id
        self._next_component_id += 1

        # Serialize the component to dict for lossless storage
        comp_dict = _component_to_dict(comp)
        comp_json = json.dumps(comp_dict)

        # Extract flat query columns
        value_raw: Any = None
        value_type: str | None = None
        if comp.value is not None:
            vt_name = get_value_type_name(comp.value)
            if vt_name is not None:
                value_type = vt_name
                if isinstance(comp.value, tuple):
                    value_raw = [int(comp.value[0]), int(comp.value[1])]
                else:
                    value_raw = int(comp.value)
            else:
                value_raw = comp.value

        tone_function = _get_prop_str(comp, "tone_function")

        comp_type = "component"
        if isinstance(comp, AbsentComponent):
            comp_type = "absent_component"
        elif isinstance(comp, UndefinedComponent):
            comp_type = "undefined_component"

        self._rows.append(
            {
                "component_id": cid,
                "ohr_id": ohr_id,
                "position": position,
                "is_reference": is_reference,
                "value": value_raw,
                "value_type": value_type,
                "component_type": comp_type,
                "status": comp.status.value,
                "tone_function": tone_function,
                "component_json": comp_json,  # lossless backup
            }
        )

    # ── DataFrame access ──

    def to_dataframe(self) -> pd.DataFrame:
        """Return the flat component table as a DataFrame."""
        if not self._rows:
            return pd.DataFrame(
                columns=[
                    "component_id",
                    "ohr_id",
                    "position",
                    "is_reference",
                    "value",
                    "value_type",
                    "component_type",
                    "status",
                    "tone_function",
                    "component_json",
                ]
            )
        return pd.DataFrame(self._rows)

    def ohr_meta_dataframe(self) -> pd.DataFrame:
        """Return OHR-level metadata as a DataFrame."""
        rows = []
        for meta in self._ohr_meta.values():
            rows.append(
                {
                    "ohr_id": meta.ohr_id,
                    "parent_ohr_id": meta.parent_ohr_id,
                    "body_position": meta.body_position,
                    "ordered": meta.ordered,
                    "horizontal": meta.horizontal,
                    "reference_ohr_id": meta.reference_ohr_id,
                    "properties_json": meta.properties_json,
                }
            )
        return (
            pd.DataFrame(rows)
            if rows
            else pd.DataFrame(
                columns=[
                    "ohr_id",
                    "parent_ohr_id",
                    "body_position",
                    "ordered",
                    "horizontal",
                    "reference_ohr_id",
                    "properties_json",
                ]
            )
        )

    # ── Reconstruction ──

    def reconstruct_ohr(self, ohr_id: int, _df: pd.DataFrame | None = None) -> OHR:
        """Reconstruct a full OHR tree from the flat table."""
        meta = self._ohr_meta[ohr_id]
        if _df is None:
            _df = self.to_dataframe()

        # Reference component
        ref_rows = _df[(_df["ohr_id"] == ohr_id) & (_df["is_reference"])]
        ref_comp: Component | None = None
        if len(ref_rows) > 0:
            ref_comp = _component_from_dict(json.loads(ref_rows.iloc[0]["component_json"]))

        # Reference OHR
        ref_ohr: OHR | None = None
        if meta.reference_ohr_id is not None:
            ref_ohr = self.reconstruct_ohr(meta.reference_ohr_id, _df=_df)

        # Body: interleave components and child OHRs by position
        body_comps = _df[(_df["ohr_id"] == ohr_id) & (~_df["is_reference"])]

        # Child OHRs that are body items (have body_position != None)
        child_body_ohrs: list[tuple[int, int]] = []  # (position, child_ohr_id)
        for m in self._ohr_meta.values():
            if m.parent_ohr_id == ohr_id and m.ohr_id != (meta.reference_ohr_id or -1):
                if m.body_position is not None:
                    child_body_ohrs.append((m.body_position, m.ohr_id))

        # Build position -> item mapping
        body_items: dict[int, Any] = {}
        for _, row in body_comps.iterrows():
            pos = row["position"]
            body_items[pos] = _component_from_dict(json.loads(row["component_json"]))

        for pos, child_id in child_body_ohrs:
            body_items[pos] = self.reconstruct_ohr(child_id, _df=_df)

        # Assemble body in position order
        body = [body_items[k] for k in sorted(body_items.keys())]

        # Reconstruct OHR-level properties
        ohr_props_data = json.loads(meta.properties_json)
        ohr_props = _properties_from_json(ohr_props_data) if ohr_props_data else EMPTY_PROPS

        return OHR(
            reference_component=ref_comp,
            reference_ohr=ref_ohr,
            body=tuple(body),
            ordered=meta.ordered,
            horizontal=meta.horizontal,
            properties=ohr_props,
        )

    def reconstruct_all_top_level(self) -> list[OHR]:
        """Reconstruct all top-level OHRs (those with no parent)."""
        # Pre-build indices for fast reconstruction
        comp_index, child_index = self._build_indices()
        top_ids = [m.ohr_id for m in self._ohr_meta.values() if m.parent_ohr_id is None]
        return [self._reconstruct_fast(oid, comp_index, child_index) for oid in sorted(top_ids)]

    def _build_indices(self) -> tuple[dict, dict]:
        """Build lookup indices for fast reconstruction.

        Returns (comp_index, child_index) where:
        - comp_index[ohr_id] = list of (position, is_reference, component_json)
        - child_index[parent_ohr_id] = list of (body_position, child_ohr_id)
        """
        comp_index: dict[int, list[tuple[int, bool, str]]] = {}
        for row in self._rows:
            oid = row["ohr_id"]
            comp_index.setdefault(oid, []).append((row["position"], row["is_reference"], row["component_json"]))

        child_index: dict[int, list[tuple[int, int]]] = {}
        for meta in self._ohr_meta.values():
            if meta.parent_ohr_id is not None and meta.body_position is not None:
                parent = meta.parent_ohr_id
                # Exclude reference_ohr children
                parent_meta = self._ohr_meta.get(parent)
                if parent_meta and parent_meta.reference_ohr_id == meta.ohr_id:
                    continue
                child_index.setdefault(parent, []).append((meta.body_position, meta.ohr_id))

        return comp_index, child_index

    def _reconstruct_fast(
        self,
        ohr_id: int,
        comp_index: dict,
        child_index: dict,
    ) -> OHR:
        """Fast OHR reconstruction using pre-built indices."""
        meta = self._ohr_meta[ohr_id]

        # Reference component
        ref_comp: Component | None = None
        body_items: dict[int, Any] = {}

        for pos, is_ref, cjson in comp_index.get(ohr_id, []):
            comp = _component_from_dict(json.loads(cjson))
            if is_ref:
                ref_comp = comp
            else:
                body_items[pos] = comp

        # Reference OHR
        ref_ohr: OHR | None = None
        if meta.reference_ohr_id is not None:
            ref_ohr = self._reconstruct_fast(meta.reference_ohr_id, comp_index, child_index)

        # Child OHRs in body
        for pos, child_id in child_index.get(ohr_id, []):
            body_items[pos] = self._reconstruct_fast(child_id, comp_index, child_index)

        body = [body_items[k] for k in sorted(body_items.keys())]

        ohr_props_data = json.loads(meta.properties_json)
        ohr_props = _properties_from_json(ohr_props_data) if ohr_props_data else EMPTY_PROPS

        return OHR(
            reference_component=ref_comp,
            reference_ohr=ref_ohr,
            body=tuple(body),
            ordered=meta.ordered,
            horizontal=meta.horizontal,
            properties=ohr_props,
        )

    # ── Queries ──

    def query_components_by_property(self, key: str, value: Any) -> pd.DataFrame:
        """Find all components matching a flat property column value."""
        df = self.to_dataframe()
        if key in df.columns:
            return df[df[key] == value]
        return df.iloc[0:0]

    def query_ohrs_by_property(self, key: str, value: Any) -> list[int]:
        """Find all OHR IDs whose metadata properties match."""
        results = []
        for meta in self._ohr_meta.values():
            props_data = json.loads(meta.properties_json)
            for prop in props_data:
                if prop.get("type") == "enum" and prop.get("value") == value:
                    results.append(meta.ohr_id)
                    break
                if prop.get("type") == "key_value" and prop.get("key") == key and prop.get("value") == value:
                    results.append(meta.ohr_id)
                    break
        return results

    def count_components(self) -> int:
        """Total number of component rows."""
        return len(self._rows)

    def count_ohrs(self) -> int:
        """Total number of OHR nodes."""
        return len(self._ohr_meta)

    @property
    def memory_bytes(self) -> int:
        """Approximate memory usage."""
        import sys

        df = self.to_dataframe()
        return df.memory_usage(deep=True).sum() + sys.getsizeof(self._ohr_meta)


def _get_prop_str(comp: Component, key: str) -> str | None:
    """Extract a string property value from a component."""
    val = comp.get_property(key)
    if val is None:
        return None
    if hasattr(val, "value"):
        return val.value
    return str(val)



"""FlexOHR base error hierarchy.

Paradigm modules register their own errors via ``register_paradigm_errors()``.
All FlexOHR exceptions multi-inherit from the corresponding built-in type
so that ``except TypeError`` etc. still catches them.
"""

from __future__ import annotations


class FlexohrError(Exception):
    """Base exception for all FlexOHR errors."""

    default_message: str = "FlexOHR error"

    def __init__(self, message: str | None = None) -> None:
        self.detail = message
        msg = f"{self.default_message}: {message}" if message else self.default_message
        super().__init__(msg)


class FlexohrTypeError(FlexohrError, TypeError):
    default_message = "FlexOHR type error"


class FlexohrValueError(FlexohrError, ValueError):
    default_message = "FlexOHR value error"


class InvalidComponentValue(FlexohrValueError):
    default_message = "Invalid component value"


class ImmutabilityError(FlexohrError, AttributeError):
    default_message = "Cannot modify frozen FlexOHR object"


class ValidationError(FlexohrError):
    default_message = "Validation failed"


class CodecRegistryError(FlexohrError, KeyError):
    default_message = "Codec registry lookup failed"

    def __init__(self, message: str | None = None) -> None:  # noqa: B042
        # KeyError.__init__ expects a single key arg; route through FlexohrError
        FlexohrError.__init__(self, message)


# ── Paradigm registration ──

_paradigm_errors: dict[str, list[type[FlexohrError]]] = {}


def register_paradigm_errors(paradigm_name: str, *error_classes: type[FlexohrError]) -> None:
    """Register paradigm-specific error classes for discovery."""
    _paradigm_errors.setdefault(paradigm_name, []).extend(error_classes)


def get_paradigm_errors(paradigm_name: str) -> list[type[FlexohrError]]:
    """Retrieve registered error classes for a paradigm."""
    return list(_paradigm_errors.get(paradigm_name, []))



from typing import Tuple, TypeVar, Union

# Point, Vector, and AbstractNumber are defined in flexohr.core.abstract
# and re-exported here for backward compatibility within pitchspace.
from flexohr.core.abstract import AbstractNumber, Point, Vector

from .errors import (
    InitValueNotConvertibleToFifths,
    InitValueNotConvertibleToSemitones,
    InitValueNotConvertibleToStep,
    OperationUndefined,
    PropertyUndefined,
)
from .utils import (
    fifths2semitones,
    fifths2steps,
    name2fifths,
    name2semitones,
    name2steps,
)


class PotentialPoint(Point):
    """A Point or a promise of a Point, that is, a vector that can yield a point once added to another point."""


class PitchClass(Point):
    @property
    def semitones(self) -> int:
        raise PropertyUndefined(type(self).__name__)

    @property
    def fifths(self) -> int:
        raise PropertyUndefined(type(self).__name__)

    @property
    def steps(self) -> int:
        raise PropertyUndefined(type(self).__name__)


class IntervalClass(Vector):
    @property
    def semitones(self) -> int:
        raise PropertyUndefined(type(self).__name__)

    @property
    def fifths(self) -> int:
        raise PropertyUndefined(type(self).__name__)

    @property
    def steps(self) -> int:
        raise PropertyUndefined(type(self).__name__)


T = TypeVar("T", bound="IntType")


class IntType(int, AbstractNumber):
    """Abstract class for custom integer types. For the usual numerical operators, instances behave
    like an integer when the other value is also an integer. If the other value, however, is an
    IntType, an NotImplementedError is thrown unless the operation is defined for the respective subclass of IntType.
    The actual int value is stored in the field _value. The various subclasses provide different properties for
    accessing this field.
    """

    @staticmethod
    def convert_init_value(value: Union[int, str]) -> int:
        """Used by the constructor to convert the initial value into the integer that will be stored."""
        return int(value)

    def __new__(cls, value: Union[int, str]) -> T:
        """Can be created from an integer or any type that int() accepts."""
        converted_value = cls.convert_init_value(value)
        instance = int.__new__(cls, converted_value)
        instance._value = converted_value
        return instance

    @property
    def name(self) -> str:
        return str(self._value)

    def __str__(self) -> str:
        """String representation should work as a constructor."""
        return f"{type(self).__name__}({self.name})"

    def __repr__(self) -> str:
        """Representation should be short for slim display within containers."""
        return f"{type(self).__name__}({self.name})"

    def __neg__(self) -> T:
        raise OperationUndefined(f"- {type(self)}")

    def __add__(self, other: Union[T, int]) -> T:
        if isinstance(other, (IntType, PairOfInts)):
            raise OperationUndefined(f"{type(self)} + {type(other)}")
        return self.__class__(self._value + other)

    def __sub__(self, other: Union[T, int]) -> T:
        if isinstance(other, (IntType, PairOfInts)):
            raise OperationUndefined(f"{type(self)} - {type(other)}")
        return self.__class__(self._value - other)

    def __mul__(self, other: Union[T, int]) -> T:
        if isinstance(other, (IntType, PairOfInts)):
            raise OperationUndefined(f"{type(self)} * {type(other)}")
        return self.__class__(self._value * other)

    def __floordiv__(self, other: Union[T, int]) -> T:
        if isinstance(other, (IntType, PairOfInts)):
            raise OperationUndefined(f"{type(self)} // {type(other)}")
        return self.__class__(self._value // other)

    def __truediv__(self, other: Union[T, int]) -> T:
        if isinstance(other, (IntType, PairOfInts)):
            raise OperationUndefined(f"{type(self)} / {type(other)}")
        return self.__class__(int(self._value / other))

    def __mod__(self, other: Union[T, int]) -> T:
        if isinstance(other, (IntType, PairOfInts)):
            raise OperationUndefined(f"{type(self)} % {type(other)}")
        return self.__class__(self._value % other)

    def __divmod__(self, other: Union[T, int]) -> Tuple[int, int]:
        if isinstance(other, (IntType, PairOfInts)):
            raise OperationUndefined(f"divmod({type(self)}, {type(other)})")
        return int(self._value).__divmod__(other)

    def __radd__(self, other: Union[T, int]) -> T:
        if isinstance(other, (IntType, PairOfInts)):
            raise OperationUndefined(f"{type(other)} + {type(self)}")
        return self.__class__(other + self._value)

    def __rsub__(self, other: Union[T, int]) -> T:
        if isinstance(other, (IntType, PairOfInts)):
            raise OperationUndefined(f"{type(other)} - {type(self)}")
        return self.__class__(other - self._value)

    def __rmul__(self, other: Union[T, int]) -> T:
        if isinstance(other, (IntType, PairOfInts)):
            raise OperationUndefined(f"{type(other)} * {type(self)}")
        return self.__class__(other * self._value)

    def __rfloordiv__(self, other: Union[T, int]) -> T:
        if isinstance(other, (IntType, PairOfInts)):
            raise OperationUndefined(f"{type(other)} // {type(self)}")
        return self.__class__(other // self._value)

    def __rtruediv__(self, other: Union[T, int]) -> T:
        if isinstance(other, (IntType, PairOfInts)):
            raise OperationUndefined(f"{type(other)} / {type(self)}")
        return self.__class__(int(other / self._value))

    def __rmod__(self, other: Union[T, int]) -> T:
        if isinstance(other, (IntType, PairOfInts)):
            raise OperationUndefined(f"{type(other)} % {type(self)}")
        return self.__class__(other % self._value)

    def __rdivmod__(self, other: Union[T, int]) -> Tuple[int, int]:
        if isinstance(other, (IntType, PairOfInts)):
            raise OperationUndefined(f"divmod({type(other)}, {type(self)})")
        return int(other).__divmod__(self._value)

    def __pow__(self, other: Union[T, int]) -> T:
        if isinstance(other, (IntType, PairOfInts)):
            raise OperationUndefined(f"{type(self)} ** {type(other)}")
        return self.__class__(self._value**other)

    def __rpow__(self, other: Union[T, int]) -> T:
        if isinstance(other, (IntType, PairOfInts)):
            raise OperationUndefined(f"{type(other)} ** {type(self)}")
        return self.__class__(other**self._value)

    def __and__(self, other: Union[T, int]) -> T:
        if isinstance(other, (IntType, PairOfInts)):
            raise OperationUndefined(f"{type(self)} & {type(other)}")
        return self.__class__(self._value & other)

    def __or__(self, other: Union[T, int]) -> T:
        if isinstance(other, (IntType, PairOfInts)):
            raise OperationUndefined(f"{type(self)} | {type(other)}")
        return self.__class__(self._value | other)

    def __xor__(self, other: Union[T, int]) -> T:
        if isinstance(other, (IntType, PairOfInts)):
            raise OperationUndefined(f"{type(self)} ^ {type(other)}")
        return self.__class__(self._value ^ other)

    def __lshift__(self, other: Union[T, int]) -> T:
        if isinstance(other, (IntType, PairOfInts)):
            raise OperationUndefined(f"{type(self)} << {type(other)}")
        return self.__class__(self._value << other)

    def __rshift__(self, other: Union[T, int]) -> T:
        if isinstance(other, (IntType, PairOfInts)):
            raise OperationUndefined(f"{type(self)} >> {type(other)}")
        return self.__class__(self._value >> other)

    def __rand__(self, other: Union[T, int]) -> T:
        if isinstance(other, (IntType, PairOfInts)):
            raise OperationUndefined(f"{type(other)} & {type(self)}")
        return self.__class__(other & self._value)

    def __ror__(self, other: Union[T, int]) -> T:
        if isinstance(other, (IntType, PairOfInts)):
            raise OperationUndefined(f"{type(other)} | {type(self)}")
        return self.__class__(other | self._value)

    def __rxor__(self, other: Union[T, int]) -> T:
        if isinstance(other, (IntType, PairOfInts)):
            raise OperationUndefined(f"{type(other)} ^ {type(self)}")
        return self.__class__(other ^ self._value)

    def __rlshift__(self, other: Union[T, int]) -> T:
        if isinstance(other, (IntType, PairOfInts)):
            raise OperationUndefined(f"{type(other)} << {type(self)}")
        return self.__class__(other << self._value)

    def __rrshift__(self, other: Union[T, int]) -> T:
        if isinstance(other, (IntType, PairOfInts)):
            raise OperationUndefined(f"{type(other)} >> {type(self)}")
        return self.__class__(other >> self._value)

    def __eq__(self, other) -> bool:
        if isinstance(other, IntType) and not (type(self) is type(other)):
            raise OperationUndefined(f"{type(self)} == {type(other)}")
        return self._value == other

    def __lt__(self, other: Union[T, int]) -> bool:
        if isinstance(other, IntType) and not (type(self) is type(other)):
            raise OperationUndefined(f"{type(self)} < {type(other)}")
        return self._value < other

    def __le__(self, other: Union[T, int]) -> bool:
        if isinstance(other, IntType) and not (type(self) is type(other)):
            raise OperationUndefined(f"{type(self)} <= {type(other)}")
        return self._value <= other

    def __gt__(self, other: Union[T, int]) -> bool:
        if isinstance(other, IntType) and not (type(self) is type(other)):
            raise OperationUndefined(f"{type(self)} > {type(other)}")
        return self._value > other

    def __ge__(self, other: Union[T, int]) -> bool:
        if isinstance(other, IntType) and not (type(self) is type(other)):
            raise OperationUndefined(f"{type(self)} >= {type(other)}")
        return self._value >= other


class Octaves(IntType):
    """Integer type whose _value field represents a number of octaves
    for combination with all types of pitch and interval classes."""

    @property
    def semitones(self) -> int:
        """This scalar's value."""
        return

    @property
    def fifths(self) -> int:
        """Semitones expressed as the number of fifths. Will require inference logic or additional information."""
        return

    @property
    def octaves(self) -> int:
        """Octave information will depend on the subtype."""
        return self._value

    @property
    def steps(self) -> int:
        return

    def __add__(self, other) -> int:
        if isinstance(other, Octaves):
            return self._value + other._value
        return super().__add__(other)

    def __sub__(self, other) -> int:
        if isinstance(other, Octaves):
            return self._value - other._value
        return super().__add__(other)

    def __neg__(self):
        return self.__new__(-self._value)


class OctaveNumber(Octaves):
    """This :class:`Octaves` integer represents a particular octave where 4 stands for the
    "middle" octave (the pitches of which would be marked with a prime ' in Helmholtz notation).
    In this convention, MIDI number 0 has OctaveNumber -1.
    """

    def __repr__(self):
        return f"Octave({self.name})"

    def __str__(self):
        return f"Octave({self.name})"


semitones = TypeVar("semitones", bound="SemitonesScalar")


class SemitonesScalar(IntType):
    """Integer type whose _value field represents a number of semitones,
    such as a MIDI number or chromatic pitch class."""

    @staticmethod
    def convert_init_value(value: Union[int, str, IntType]) -> int:
        if isinstance(value, FifthsScalar):
            converted = fifths2semitones(value)
        elif isinstance(value, StepsScalar):
            raise TypeError(f"Cannot convert {type(value)} to semitones.")
        elif isinstance(value, str):
            converted = name2semitones(value)
        else:
            converted = int(value)
        return converted % 12

    @property
    def semitones(self) -> int:
        """This scalar's value."""
        return self._value

    @property
    def fifths(self) -> int:
        """Semitones expressed as the number of fifths. Will require inference logic or additional information."""
        raise PropertyUndefined(type(self).__name__)

    @property
    def octave(self) -> int:
        """OctaveNumber information will depend on the subtype."""
        raise PropertyUndefined(type(self).__name__)

    @property
    def octaves(self) -> int:
        """Octave information will depend on the subtype."""
        raise PropertyUndefined(type(self).__name__)

    def __new__(cls, value: Union[int, str]) -> semitones:
        try:
            return super().__new__(cls, value)
        except (TypeError, ValueError):
            raise InitValueNotConvertibleToSemitones(cls=cls, val=value)

    def __neg__(self) -> int:
        return super().__neg__()


fifths = TypeVar("fifths", bound="FifthsScalar")


class FifthsScalar(IntType):
    """Integer type whose _value field represents a number of perfect fifth intervals,
    such as a spelled pitch class or a specific interval."""

    @staticmethod
    def convert_init_value(value: Union[int, str, IntType]) -> int:
        if isinstance(value, str):
            converted = name2fifths(value)
        elif isinstance(value, Union[SemitonesScalar, StepsScalar]):
            raise TypeError(f"Cannot convert {type(value)} to fifths.")
        else:
            converted = int(value)
        return converted

    @property
    def fifths(self) -> int:
        """This scalar's value."""
        return self._value

    @property
    def semitones(self) -> int:
        """The stack of fifths represented by this scalar, expressed in semitones."""
        return fifths2semitones(self._value)

    @property
    def steps(self) -> int:
        return fifths2steps(self._value)

    @property
    def octave(self) -> int:
        """OctaveNumber information will depend on the subtype."""
        raise PropertyUndefined(type(self).__name__)

    @property
    def octaves(self) -> int:
        """Octave information will depend on the subtype."""
        raise PropertyUndefined(type(self).__name__)

    def __new__(cls, value: Union[int, str]) -> fifths:
        try:
            return super().__new__(cls, value)
        except (TypeError, ValueError):
            raise InitValueNotConvertibleToFifths(cls=cls, val=value)

    def __str__(self) -> str:
        """Constructors of subtypes will work with string representations."""
        return f"{type(self).__name__}('{self.name}')"

    def __repr__(self) -> str:
        return self.name

    def __neg__(self) -> int:
        return super().__neg__()


steps = TypeVar("steps", bound="StepsScalar")


class StepsScalar(IntType):
    """Integer type whose _value field represents a number of steps,
    such as a generic pitch class or a generic interval."""

    @staticmethod
    def convert_init_value(value: Union[int, str, IntType]) -> int:

        if isinstance(value, str):
            converted = name2steps(value)
        elif isinstance(value, SemitonesScalar):
            raise TypeError(f"Cannot convert {type(value)} to steps.")
        elif isinstance(value, FifthsScalar):
            converted = fifths2steps(value)
        else:
            converted = int(value)
        return converted % 7

    def __new__(cls, value: Union[int, str]) -> fifths:
        try:
            return super().__new__(cls, value)
        except (TypeError, ValueError):
            raise InitValueNotConvertibleToStep(cls=cls, val=value)

    def __str__(self) -> str:
        """Constructors of subtypes will work with string representations."""
        return f"{type(self).__name__}('{self.name}')"

    def __repr__(self) -> str:
        return self.name

    def __neg__(self) -> int:
        return super().__neg__()

    @property
    def fifths(self) -> int:
        raise PropertyUndefined(type(self).__name__)

    @property
    def semitones(self) -> int:
        raise PropertyUndefined(type(self).__name__)

    @property
    def steps(self) -> int:
        return self._value

    @property
    def octave(self) -> int:
        """OctaveNumber information will depend on the subtype."""
        raise PropertyUndefined(type(self).__name__)

    @property
    def octaves(self) -> int:
        """Octave information will depend on the subtype."""
        raise PropertyUndefined(type(self).__name__)


U = TypeVar("U", bound=IntType)
P = TypeVar("P", bound="PairOfInts")


class PairOfInts(Tuple[T, U]):
    """Abstract class representing a pair (tuple with 2 elements) of IntType instances. Its purpose is to
    bundle functionality for all pitches and intervals, which are being represented as combinations of
    pitch/interval class and an Octaves values.
    Mirroring the IntType implementation, the usual numerical operators operate on the left element (the
    pitch/interval class) when `other` is an integer or IntType, leaving the right element (the Octaves) untouched.
    If the other value, however, is another PairOfInts or a tuple of integers,
    an NotImplementedError is thrown unless the operation is defined for the respective subclass of PairOfInts.
    """

    @staticmethod
    def convert_init_value(
        value: Union[Tuple[T | int, U | int], str, int, Tuple[Tuple]],
    ) -> Tuple[T, U]:
        """Used by the constructor to convert the initial value into a pair of IntType."""
        if isinstance(value, str):
            raise NotImplementedError
        if isinstance(value, int):
            raise NotImplementedError
        a, b = value
        if not isinstance(a, IntType):
            a = IntType(a)
        if not isinstance(b, IntType):
            b = IntType(b)
        return (a, b)

    def __new__(cls, value: Union[Tuple[T | int, U | int], str, int]) -> P:
        if type(value) is cls:
            converted_value = value
        else:
            converted_value = cls.convert_init_value(value)
        instance = tuple.__new__(cls, converted_value)
        instance._value = converted_value
        return instance

    @property
    def name(self) -> str:
        return f"{self[0].name}{self[1].name}"

    def __str__(self) -> str:
        """String representation should work as a constructor."""
        return f"{type(self).__name__}({self.name})"

    def __repr__(self) -> str:
        """Representation should be short for slim display within containers."""
        return f"{type(self).__name__}({self.name})"

    def __neg__(self) -> Tuple[T, U]:
        raise OperationUndefined(f"- {type(self)}")

    def __add__(self, other: Union[T, int, P, Tuple[T | int, U | int]]) -> Tuple[T, U]:
        if isinstance(other, PairOfInts):
            raise OperationUndefined(f"{type(self)} + {type(other)}")
        if isinstance(other, tuple):
            t, u = other
            return self.__class__((self[0] + t, self[1] + u))
        return self.__class__((self[0] + other, self[1]))

    def __sub__(self, other: Union[T, int, P, Tuple[T | int, U | int]]) -> Tuple[T, U]:
        if isinstance(other, PairOfInts):
            raise OperationUndefined(f"{type(self)} - {type(other)}")
        if isinstance(other, tuple):
            t, u = other
            return self.__class__((self[0] - t, self[1] - u))
        return self.__class__((self[0] - other, self[1]))

    def __mul__(self, other: Union[T, int, P, Tuple[T | int, U | int]]) -> Tuple[T, U]:
        if isinstance(other, (PairOfInts, tuple)):
            raise OperationUndefined(f"{type(self)} * {type(other)}")
        return self.__class__((self[0] * other, self[1]))

    def __floordiv__(self, other: Union[T, int, P, Tuple[T | int, U | int]]) -> Tuple[T, U]:
        if isinstance(other, (PairOfInts, tuple)):
            raise OperationUndefined(f"{type(self)} // {type(other)}")
        return self.__class__((self[0] // other, self[1]))

    def __truediv__(self, other: Union[T, int, P, Tuple[T | int, U | int]]) -> Tuple[T, U]:
        if isinstance(other, (PairOfInts, tuple)):
            raise OperationUndefined(f"{type(self)} / {type(other)}")
        return self.__class__((self[0] / other, self[1]))

    def __mod__(self, other: Union[T, int, P, Tuple[T | int, U | int]]) -> Tuple[T, U]:
        if isinstance(other, (PairOfInts, tuple)):
            raise OperationUndefined(f"{type(self)} % {type(other)}")
        return self.__class__((self[0] % other, self[1]))

    def __divmod__(self, other: Union[T, int, P, Tuple[T | int, U | int]]) -> Tuple[int, int]:
        raise OperationUndefined(f"divmod({type(self)}, {type(other)})")

    def __radd__(self, other: Union[T, int, P, Tuple[T | int, U | int]]) -> Tuple[T, U]:
        if isinstance(other, PairOfInts):
            raise OperationUndefined(f"{type(self)} + {type(other)}")
        if isinstance(other, tuple):
            t, u = other
            return self.__class__((t + self[0], u + self[1]))
        return self.__class__((other + self[0], self[1]))

    def __rsub__(self, other: Union[T, int, P, Tuple[T | int, U | int]]) -> Tuple[T, U]:
        if isinstance(other, PairOfInts):
            raise OperationUndefined(f"{type(self)} - {type(other)}")
        if isinstance(other, tuple):
            t, u = other
            return self.__class__((t - self[0], u - self[1]))
        return self.__class__((other - self[0], self[1]))

    def __rmul__(self, other: Union[T, int, P, Tuple[T | int, U | int]]) -> Tuple[T, U]:
        if isinstance(other, (PairOfInts, tuple)):
            raise OperationUndefined(f"{type(self)} * {type(other)}")
        return self.__class__((other * self[0], self[1]))

    def __rfloordiv__(self, other: Union[T, int, P, Tuple[T | int, U | int]]) -> Tuple[T, U]:
        if isinstance(other, (PairOfInts, tuple)):
            raise OperationUndefined(f"{type(self)} // {type(other)}")
        return self.__class__((other // self[0], self[1]))

    def __rtruediv__(self, other: Union[T, int, P, Tuple[T | int, U | int]]) -> Tuple[T, U]:
        if isinstance(other, (PairOfInts, tuple)):
            raise OperationUndefined(f"{type(self)} / {type(other)}")
        return self.__class__((other / self[0], self[1]))

    def __rmod__(self, other: Union[T, int, P, Tuple[T | int, U | int]]) -> Tuple[T, U]:
        if isinstance(other, (PairOfInts, tuple)):
            raise OperationUndefined(f"{type(self)} % {type(other)}")
        return self.__class__((other % self[0], self[1]))

    def __rdivmod__(self, other: Union[T, int]) -> Tuple[int, int]:
        raise OperationUndefined(f"divmod({type(other)}, {type(self)})")

    def __pow__(self, other: Union[T, int, P, Tuple[T | int, U | int]]) -> Tuple[T, U]:
        if isinstance(other, (PairOfInts, tuple)):
            raise OperationUndefined(f"{type(self)} ** {type(other)}")
        return self.__class__((self[0] ** other, self[1]))

    def __rpow__(self, other: Union[T, int, P, Tuple[T | int, U | int]]) -> Tuple[T, U]:
        if isinstance(other, (PairOfInts, tuple)):
            raise OperationUndefined(f"{type(other)} ** {type(self)}")
        return self.__class__((other ** self[0], self[1]))

    def __and__(self, other: Union[T, int, P, Tuple[T | int, U | int]]) -> Tuple[T, U]:
        if isinstance(other, (PairOfInts, tuple)):
            raise OperationUndefined(f"{type(self)} & {type(other)}")
        return self.__class__((self[0] & other, self[1]))

    def __or__(self, other: Union[T, int, P, Tuple[T | int, U | int]]) -> Tuple[T, U]:
        if isinstance(other, (PairOfInts, tuple)):
            raise OperationUndefined(f"{type(self)} | {type(other)}")
        return self.__class__((self[0] | other, self[1]))

    def __xor__(self, other: Union[T, int, P, Tuple[T | int, U | int]]) -> Tuple[T, U]:
        if isinstance(other, (PairOfInts, tuple)):
            raise OperationUndefined(f"{type(self)} ^ {type(other)}")
        return self.__class__((self[0] ^ other, self[1]))

    def __lshift__(self, other: Union[T, int, P, Tuple[T | int, U | int]]) -> Tuple[T, U]:
        if isinstance(other, (PairOfInts, tuple)):
            raise OperationUndefined(f"{type(self)} << {type(other)}")
        return self.__class__((self[0] << other, self[1]))

    def __rshift__(self, other: Union[T, int, P, Tuple[T | int, U | int]]) -> Tuple[T, U]:
        if isinstance(other, (PairOfInts, tuple)):
            raise OperationUndefined(f"{type(self)} >> {type(other)}")
        return self.__class__((self[0] >> other, self[1]))

    def __rand__(self, other: Union[T, int, P, Tuple[T | int, U | int]]) -> Tuple[T, U]:
        if isinstance(other, (PairOfInts, tuple)):
            raise OperationUndefined(f"{type(other)} & {type(self)}")
        return self.__class__((other & self[0], self[1]))

    def __ror__(self, other: Union[T, int, P, Tuple[T | int, U | int]]) -> Tuple[T, U]:
        if isinstance(other, (PairOfInts, tuple)):
            raise OperationUndefined(f"{type(other)} | {type(self)}")
        return self.__class__((other | self[0], self[1]))

    def __rxor__(self, other: Union[T, int, P, Tuple[T | int, U | int]]) -> Tuple[T, U]:
        if isinstance(other, (PairOfInts, tuple)):
            raise OperationUndefined(f"{type(other)} ^ {type(self)}")
        return self.__class__((other ^ self[0], self[1]))

    def __rlshift__(self, other: Union[T, int, P, Tuple[T | int, U | int]]) -> Tuple[T, U]:
        if isinstance(other, (PairOfInts, tuple)):
            raise OperationUndefined(f"{type(other)} << {type(self)}")
        return self.__class__((other << self[0], self[1]))

    def __rrshift__(self, other: Union[T, int, P, Tuple[T | int, U | int]]) -> Tuple[T, U]:
        if isinstance(other, (PairOfInts, tuple)):
            raise OperationUndefined(f"{type(other)} >> {type(self)}")
        return self.__class__((other >> self[0], self[1]))

    def __eq__(self, other: Union[T, int, P, Tuple[T | int, U | int]]) -> bool:
        if isinstance(other, int) or (isinstance(other, PairOfInts) and not isinstance(self, type(other))):
            raise OperationUndefined(f"{type(self)} == {type(other)}")
        return self == other

    def __lt__(self, other: Union[T, int, P, Tuple[T | int, U | int]]) -> bool:
        if isinstance(other, int) or (isinstance(other, PairOfInts) and not isinstance(self, type(other))):
            raise OperationUndefined(f"{type(self)} < {type(other)}")
        return self < other

    def __le__(self, other: Union[T, int, P, Tuple[T | int, U | int]]) -> bool:
        if isinstance(other, int) or (isinstance(other, PairOfInts) and not isinstance(self, type(other))):
            raise OperationUndefined(f"{type(self)} <= {type(other)}")
        return self <= other

    def __gt__(self, other: Union[T, int, P, Tuple[T | int, U | int]]) -> bool:
        if isinstance(other, int) or (isinstance(other, PairOfInts) and not isinstance(self, type(other))):
            raise OperationUndefined(f"{type(self)} > {type(other)}")
        return self > other

    def __ge__(self, other: Union[T, int, P, Tuple[T | int, U | int]]) -> bool:
        if isinstance(other, int) or (isinstance(other, PairOfInts) and not isinstance(self, type(other))):
            raise OperationUndefined(f"{type(self)} >= {type(other)}")
        return self >= other


Pi = TypeVar("Pi", bound=PitchClass)
Ocs = TypeVar("Ocs", bound=Octaves)
Oc = TypeVar("Oc", bound=OctaveNumber)
Iv = TypeVar("Iv", bound=IntervalClass)


class Pitch(PairOfInts[Pi, Oc], Point):
    @property
    def octave(self) -> Oc:
        return self[1]

    @property
    def pitch_class(self) -> Pi:
        return self[0]

    @property
    def semitones(self) -> int:
        raise PropertyUndefined(type(self).__name__)

    @property
    def fifths(self) -> int:
        raise PropertyUndefined(type(self).__name__)

    @property
    def steps(self) -> int:
        raise PropertyUndefined(type(self).__name__)


class Interval(PairOfInts[Iv, Ocs], Vector):
    @property
    def octaves(self) -> Ocs:
        return self[1]

    @property
    def interval_class(self) -> Iv:
        return self[0]

    @property
    def semitones(self) -> int:
        raise PropertyUndefined(type(self).__name__)

    @property
    def fifths(self) -> int:
        raise PropertyUndefined(type(self).__name__)

    @property
    def steps(self) -> int:
        raise PropertyUndefined(type(self).__name__)


if __name__ == "__main__":
    a = SemitonesScalar(4)
    b = SemitonesScalar("5")
    c = FifthsScalar(6)
    d = FifthsScalar("7")
    # these operations will return integers
    print(f"{a} + 1 = {a + 1}")
    print(f"{b} * -1 = {b * -1}")
    print(f"{c} // 2 = {c // 2}")
    print(f"{d} ** 2 = {d ** 2}")
    # these operations will fail
    try:
        a + b
    except NotImplementedError as e:
        print(f'{a} + {b} failed with "{e}"')
    try:
        b * c
    except NotImplementedError as e:
        print(f'{b} * {c} failed with "{e}"')
    try:
        -a
    except NotImplementedError as e:
        print(f'-{a} failed with "{e}"')



"""Accidental — chromatic alteration as SI subclass."""

from __future__ import annotations

from flexohr.paradigms.pitchspace.abstract import Octaves
from flexohr.paradigms.pitchspace.interval import SpecificInterval as SI
from flexohr.paradigms.pitchspace.interval import SpecificIntervalClass as SIC

_SHARP = "\u266f"  # ♯
_FLAT = "\u266d"  # ♭
_NATURAL = "\u266e"  # ♮


class Accidental(SI):
    """Chromatic alteration as a specific interval with octave 0.

    Invariant: the octave component is always ``Octaves(0)``.
    The value on the line of fifths is ``n * 7`` where *n* is the number
    of sharps (positive) or flats (negative).
    """

    @staticmethod
    def convert_init_value(value):
        if isinstance(value, SI):
            return SIC(value.interval_class.fifths), Octaves(0)
        if isinstance(value, SIC):
            return SIC(value.fifths), Octaves(0)
        if isinstance(value, tuple):
            return SIC(value[0]), Octaves(0)
        if isinstance(value, int):
            return SIC(value), Octaves(0)
        raise TypeError(
            f"Accidental cannot be constructed from {type(value).__name__}. "
            f"Use Accidental.from_int(), .from_string(), or pass an SI/SIC."
        )

    @classmethod
    def from_int(cls, n: int) -> Accidental:
        """Create from a count of sharps (positive) or flats (negative)."""
        return cls(SIC(n * 7))

    @classmethod
    def from_string(cls, s: str) -> Accidental:
        """Parse accidental string.

        Accepted: ``♯``, ``♭``, ``♮``, ``#``, ``b``, ``-`` (music21),
        and multiples (``##``, ``bb``, ``--``, ``♯♯``, ``♭♭``).
        """
        if not s or s == _NATURAL:
            return cls.from_int(0)
        if all(c in ("#", _SHARP) for c in s):
            return cls.from_int(len(s))
        if all(c in ("b", _FLAT) for c in s):
            return cls.from_int(-len(s))
        if all(c == "-" for c in s):
            return cls.from_int(-len(s))
        raise ValueError(f"Cannot parse accidental string: {s!r}")

    @property
    def n(self) -> int:
        """Count of sharps (positive) or flats (negative)."""
        return self.interval_class.fifths // 7

    def to_string(self) -> str:
        """Unicode accidental string."""
        n = self.n
        if n > 0:
            return _SHARP * n
        if n < 0:
            return _FLAT * -n
        return _NATURAL

    def __repr__(self):
        return f"acc({self.to_string()})"

    def __str__(self):
        return f"acc({self.to_string()})"

    def __hash__(self):
        return hash(("Accidental", self.interval_class.fifths))

    def __eq__(self, other):
        if isinstance(other, Accidental):
            return self.interval_class.fifths == other.interval_class.fifths
        return NotImplemented

    def __add__(self, other):
        if isinstance(other, Accidental):
            return Accidental.from_int(self.n + other.n)
        return super().__add__(other)

    def __sub__(self, other):
        if isinstance(other, Accidental):
            return Accidental.from_int(self.n - other.n)
        return super().__sub__(other)

    def __neg__(self):
        return Accidental.from_int(-self.n)

    def __radd__(self, other):
        return NotImplemented


Accidental.NATURAL = Accidental.from_int(0)
Accidental.SHARP = Accidental.from_int(1)
Accidental.FLAT = Accidental.from_int(-1)
Accidental.DOUBLE_SHARP = Accidental.from_int(2)
Accidental.DOUBLE_FLAT = Accidental.from_int(-2)
Accidental.TRIPLE_SHARP = Accidental.from_int(3)
Accidental.TRIPLE_FLAT = Accidental.from_int(-3)



"""Chord quality -> interval structure mapping.

Each chord quality maps to a tuple of ``(SIC, tone_functions)`` pairs where
``tone_functions`` is a tuple of one or more :class:`ToneFunction` members.
Standard chords have a single ToneFunction per component; augmented sixth
chords have multiple functions per component, derived from their resolution
labels.

SIC is imported from the pitchspace paradigm module.
"""

from __future__ import annotations

from flexohr.harmony.harmony_enums import (
    ChordClass,
    ChordQuality,
    CollectionType,
    Inversion,
    ToneFunction,
)
from flexohr.paradigms.pitchspace import SIC

# Using the function name `tf` as shorthand
tf = ToneFunction

# Type: dict[ChordQuality, tuple[tuple[SIC, tuple[ToneFunction, ...]], ...]]
# Each entry is (interval, (tone_function_1, tone_function_2, ...))
CHORD_QUALITY_INTERVALS: dict[ChordQuality, tuple[tuple[SIC, tuple[ToneFunction, ...]], ...]] = {
    ChordQuality.major_triad: (
        (SIC("P1"), (tf.root,)),
        (SIC("M3"), (tf.third,)),
        (SIC("P5"), (tf.fifth,)),
    ),
    ChordQuality.minor_triad: (
        (SIC("P1"), (tf.root,)),
        (SIC("m3"), (tf.third,)),
        (SIC("P5"), (tf.fifth,)),
    ),
    ChordQuality.diminished_triad: (
        (SIC("P1"), (tf.root,)),
        (SIC("m3"), (tf.third,)),
        (SIC("d5"), (tf.fifth,)),
    ),
    ChordQuality.augmented_triad: (
        (SIC("P1"), (tf.root,)),
        (SIC("M3"), (tf.third,)),
        (SIC("a5"), (tf.fifth,)),
    ),
    ChordQuality.dominant_seventh: (
        (SIC("P1"), (tf.root,)),
        (SIC("M3"), (tf.third,)),
        (SIC("P5"), (tf.fifth,)),
        (SIC("m7"), (tf.seventh,)),
    ),
    ChordQuality.minor_seventh: (
        (SIC("P1"), (tf.root,)),
        (SIC("m3"), (tf.third,)),
        (SIC("P5"), (tf.fifth,)),
        (SIC("m7"), (tf.seventh,)),
    ),
    ChordQuality.major_seventh: (
        (SIC("P1"), (tf.root,)),
        (SIC("M3"), (tf.third,)),
        (SIC("P5"), (tf.fifth,)),
        (SIC("M7"), (tf.seventh,)),
    ),
    ChordQuality.minor_major_seventh: (
        (SIC("P1"), (tf.root,)),
        (SIC("m3"), (tf.third,)),
        (SIC("P5"), (tf.fifth,)),
        (SIC("M7"), (tf.seventh,)),
    ),
    ChordQuality.diminished_seventh: (
        (SIC("P1"), (tf.root,)),
        (SIC("m3"), (tf.third,)),
        (SIC("d5"), (tf.fifth,)),
        (SIC("d7"), (tf.seventh,)),
    ),
    ChordQuality.half_diminished_seventh: (
        (SIC("P1"), (tf.root,)),
        (SIC("m3"), (tf.third,)),
        (SIC("d5"), (tf.fifth,)),
        (SIC("m7"), (tf.seventh,)),
    ),
    ChordQuality.augmented_seventh: (
        (SIC("P1"), (tf.root,)),
        (SIC("M3"), (tf.third,)),
        (SIC("a5"), (tf.fifth,)),
        (SIC("m7"), (tf.seventh,)),
    ),
    ChordQuality.augmented_major_seventh: (
        (SIC("P1"), (tf.root,)),
        (SIC("M3"), (tf.third,)),
        (SIC("a5"), (tf.fifth,)),
        (SIC("M7"), (tf.seventh,)),
    ),
    # Augmented sixths — intervals measured from the bass note (NOT the root).
    # Tone functions are derived from the resolution labels:
    #
    #   It6  resolves as viio(b3)/V:
    #     P1  = bass (= third of resolution, altered, downward leading tone)
    #     M3  = fifth of resolution
    #     a6  = root of resolution (= leading tone, resolves upward)
    #
    #   Ger6 resolves as viio65(b3)/V:
    #     P1  = bass (= third of resolution, altered, downward leading tone)
    #     M3  = fifth of resolution
    #     P5  = seventh of resolution
    #     a6  = root of resolution (= leading tone, resolves upward)
    #
    #   Fr6  resolves as V43(b5)/V:
    #     P1  = bass (= fifth of resolution, altered, downward leading tone)
    #     M3  = seventh of resolution
    #     a4  = root of resolution
    #     a6  = third of resolution (= leading tone, resolves upward)
    #
    # Default inversion is first ("6") for all three; for Ger/Fr tetrads "6" is
    # understood as this specific inversion even though it would normally not be
    # a valid figbass for tetrads.
    ChordQuality.italian_sixth: (
        (SIC("P1"), (tf.bass, tf.third, tf.altered, tf.downward_leading_tone)),
        (SIC("M3"), (tf.fifth,)),
        (SIC("a6"), (tf.root, tf.leading_tone)),
    ),
    ChordQuality.german_sixth: (
        (SIC("P1"), (tf.bass, tf.third, tf.altered, tf.downward_leading_tone)),
        (SIC("M3"), (tf.fifth,)),
        (SIC("P5"), (tf.seventh,)),
        (SIC("a6"), (tf.root, tf.leading_tone)),
    ),
    ChordQuality.french_sixth: (
        (SIC("P1"), (tf.bass, tf.fifth, tf.altered, tf.downward_leading_tone)),
        (SIC("M3"), (tf.seventh,)),
        (SIC("a4"), (tf.root,)),
        (SIC("a6"), (tf.third, tf.leading_tone)),
    ),
    # Suspended
    ChordQuality.suspended_second: (
        (SIC("P1"), (tf.root,)),
        (SIC("M2"), (tf.suspended_second,)),
        (SIC("P5"), (tf.fifth,)),
    ),
    ChordQuality.suspended_fourth: (
        (SIC("P1"), (tf.root,)),
        (SIC("P4"), (tf.suspended_fourth,)),
        (SIC("P5"), (tf.fifth,)),
    ),
    # Power chord
    ChordQuality.power_chord: (
        (SIC("P1"), (tf.root,)),
        (SIC("P5"), (tf.fifth,)),
    ),
    # Unison
    ChordQuality.unison: ((SIC("P1"), (tf.root,)),),
}


# Derive ChordQuality -> ChordClass mapping from interval count
def _derive_chord_class(quality: ChordQuality) -> ChordClass:
    from flexohr.harmony.harmony_enums import ChordClass

    if quality in (ChordQuality.italian_sixth, ChordQuality.german_sixth, ChordQuality.french_sixth):
        return ChordClass.augmented_sixth
    if quality in (ChordQuality.suspended_second, ChordQuality.suspended_fourth):
        return ChordClass.suspended
    intervals = CHORD_QUALITY_INTERVALS.get(quality)
    if intervals is None:
        return ChordClass.unclassified
    n = len(intervals)
    if n == 1:
        return ChordClass.unclassified
    if n == 2:
        return ChordClass.dyad
    if n == 3:
        return ChordClass.triad
    if n == 4:
        return ChordClass.seventh
    return ChordClass.unclassified


CHORD_QUALITY_TO_CLASS: dict[ChordQuality, ChordClass] = {
    q: _derive_chord_class(q) for q in ChordQuality if q.name == q.value  # skip aliases
}


# ── Inversion tables ──

# Maps inversion index (0-based) to the ToneFunction whose component becomes
# the bass.  Index 0 = root position (bass is root), 1 = first inversion
# (bass is third), etc.
INVERSION_INDEX_TO_TONE_FUNCTION: dict[int, ToneFunction] = {
    0: ToneFunction.root,
    1: ToneFunction.third,
    2: ToneFunction.fifth,
    3: ToneFunction.seventh,
}

# Maps Inversion enum member to 0-based index.
INVERSION_TO_INDEX: dict[Inversion, int] = {
    Inversion.root_position: 0,
    Inversion.first: 1,
    Inversion.second: 2,
    Inversion.third: 3,
}

# Maps 0-based index to Inversion enum member.
INDEX_TO_INVERSION: dict[int, Inversion] = {v: k for k, v in INVERSION_TO_INDEX.items()}

# Maximum inversion for each ChordClass that supports inversion.
# Chord classes not in this dict do NOT support inversion.
CHORD_CLASS_MAX_INVERSION: dict[ChordClass, int] = {
    ChordClass.triad: 2,
    ChordClass.seventh: 3,
    ChordClass.augmented_sixth: 3,
}


# Scale/collection interval tables
# SIC values are positions on the line of fifths (0 = P1, 1 = P5, 2 = M2, etc.)
COLLECTION_TYPE_INTERVALS: dict[CollectionType, tuple[SIC, ...]] = {
    CollectionType.major: (
        SIC(0),
        SIC(2),
        SIC(4),
        SIC(-1),
        SIC(1),
        SIC(3),
        SIC(5),
    ),
    CollectionType.natural_minor: (
        SIC(0),
        SIC(2),
        SIC(-3),
        SIC(-1),
        SIC(1),
        SIC(-4),
        SIC(-2),
    ),
    CollectionType.aeolian: (
        SIC(0),
        SIC(2),
        SIC(-3),
        SIC(-1),
        SIC(1),
        SIC(-4),
        SIC(-2),
    ),
    CollectionType.dorian: (
        SIC(0),
        SIC(2),
        SIC(-3),
        SIC(-1),
        SIC(1),
        SIC(3),
        SIC(-2),
    ),
    CollectionType.phrygian: (
        SIC(0),
        SIC(-5),
        SIC(-3),
        SIC(-1),
        SIC(1),
        SIC(-4),
        SIC(-2),
    ),
    CollectionType.lydian: (
        SIC(0),
        SIC(2),
        SIC(4),
        SIC(6),
        SIC(1),
        SIC(3),
        SIC(5),
    ),
    CollectionType.mixolydian: (
        SIC(0),
        SIC(2),
        SIC(4),
        SIC(-1),
        SIC(1),
        SIC(3),
        SIC(-2),
    ),
    CollectionType.locrian: (
        SIC(0),
        SIC(-5),
        SIC(-3),
        SIC(-1),
        SIC(-6),
        SIC(-4),
        SIC(-2),
    ),
    CollectionType.harmonic_minor: (
        SIC(0),
        SIC(2),
        SIC(-3),
        SIC(-1),
        SIC(1),
        SIC(-4),
        SIC(5),
    ),
    CollectionType.melodic_minor: (
        SIC(0),
        SIC(2),
        SIC(-3),
        SIC(-1),
        SIC(1),
        SIC(3),
        SIC(5),
    ),
    CollectionType.major_pentatonic: (
        SIC(0),
        SIC(2),
        SIC(4),
        SIC(1),
        SIC(3),
    ),
    CollectionType.minor_pentatonic: (
        SIC(0),
        SIC(-3),
        SIC(-1),
        SIC(1),
        SIC(-2),
    ),
}



"""Random OHR sampler for testing and benchmarking.

Generates random OHR trees with configurable depth, breadth, and content.
Supports pitchspace types, generic values, and chord-table-based generation.
"""

from __future__ import annotations

import logging
import random
from dataclasses import dataclass
from typing import Any

from flexohr.core.component import AbsentComponent, Component, UndefinedComponent
from flexohr.core.ohr import OHR
from flexohr.core.validators import Validator
from flexohr.harmony.enums import EMPTY_PROPS, FrozenProperties, props

logger = logging.getLogger(__name__)


@dataclass
class SamplerConfig:
    """Configuration for random OHR generation."""

    max_depth: int = 3
    max_breadth: int = 5
    min_breadth: int = 1
    use_pitchspace: bool = True
    use_generic: bool = True
    property_density: float = 0.7
    absent_probability: float = 0.1
    undefined_probability: float = 0.05
    nesting_probability: float = 0.3
    reference_probability: float = 0.6
    reference_ohr_probability: float = 0.2
    ordered_probability: float = 0.3
    horizontal_probability: float = 0.2


class OHRSampler:
    """Random OHR generator for testing and benchmarking."""

    def __init__(
        self,
        config: SamplerConfig | None = None,
        seed: int | None = None,
    ) -> None:
        self.config = config or SamplerConfig()
        self.rng = random.Random(seed)

    def sample(self) -> OHR:
        """Generate a single random OHR."""
        return self._sample_ohr(depth=0)

    def sample_batch(self, n: int) -> list[OHR]:
        """Generate *n* random OHRs."""
        return [self.sample() for _ in range(n)]

    def sample_valid(self, validator: Validator, max_attempts: int = 100) -> OHR:
        """Generate an OHR that passes *validator*.

        Raises ``RuntimeError`` if no valid OHR is found within *max_attempts*.
        """
        for _ in range(max_attempts):
            ohr = self.sample()
            if validator(ohr).is_valid:
                return ohr
        raise RuntimeError(f"Could not generate a valid OHR in {max_attempts} attempts")

    def sample_invalid(self, validator: Validator, max_attempts: int = 100) -> OHR:
        """Generate an OHR that fails *validator*.

        Raises ``RuntimeError`` if no invalid OHR is found within *max_attempts*.
        """
        for _ in range(max_attempts):
            ohr = self._sample_intentionally_invalid()
            if not validator(ohr).is_valid:
                return ohr
        raise RuntimeError(f"Could not generate an invalid OHR in {max_attempts} attempts")

    def sample_from_chord_table(self) -> OHR:
        """Generate an OHR from a random ChordQuality with a random root.

        Picks a random chord quality from ``CHORD_QUALITY_INTERVALS``, a
        random root SPC, and assembles the corresponding OHR.
        """
        from flexohr.harmony.chord_tables import CHORD_QUALITY_INTERVALS
        from flexohr.paradigms.pitchspace import SPC

        qualities = list(CHORD_QUALITY_INTERVALS.keys())
        quality = self.rng.choice(qualities)
        intervals = CHORD_QUALITY_INTERVALS[quality]

        # Random root: fifths in [-7, 7] gives common note names
        root = SPC(self.rng.randint(-7, 7))

        first_sic, first_tfs = intervals[0]
        body_components = []
        for sic, tfs in intervals[1:]:
            body_components.append(Component(sic, properties=props(*tfs)))

        return OHR(
            reference_component=Component(root, properties=props(*first_tfs)),
            body=tuple(body_components),
            properties=props(quality),
        )

    # ── Internal helpers ──

    def _sample_ohr(self, depth: int) -> OHR:
        """Recursively generate a random OHR tree."""
        cfg = self.config

        # Reference component
        ref_comp: Component | AbsentComponent | UndefinedComponent | None = None
        if self.rng.random() < cfg.reference_probability:
            ref_comp = self._sample_component()

        # Reference OHR (only if not at max depth)
        ref_ohr: OHR | None = None
        if depth + 1 < cfg.max_depth and self.rng.random() < cfg.reference_ohr_probability:
            ref_ohr = self._sample_ohr(depth + 1)

        # Body
        breadth = self.rng.randint(cfg.min_breadth, cfg.max_breadth)
        body: list[Component | OHR] = []
        for _ in range(breadth):
            # Nesting probability decreases with depth
            effective_nesting_p = cfg.nesting_probability * (1.0 - depth / max(cfg.max_depth, 1))
            if depth + 1 < cfg.max_depth and self.rng.random() < effective_nesting_p:
                body.append(self._sample_ohr(depth + 1))
            else:
                body.append(self._sample_component())

        # Flags
        ordered = self.rng.random() < cfg.ordered_probability
        horizontal = self.rng.random() < cfg.horizontal_probability

        # OHR-level properties
        ohr_props = self._sample_ohr_properties()

        return OHR(
            reference_component=ref_comp,
            reference_ohr=ref_ohr,
            body=tuple(body),
            ordered=ordered,
            horizontal=horizontal,
            properties=ohr_props,
        )

    def _sample_component(self) -> Component | AbsentComponent | UndefinedComponent:
        """Generate a random Component, AbsentComponent, or UndefinedComponent."""
        cfg = self.config
        roll = self.rng.random()

        if roll < cfg.undefined_probability:
            return UndefinedComponent(properties=self._sample_properties())

        if roll < cfg.undefined_probability + cfg.absent_probability:
            return AbsentComponent(
                value=self._sample_value(),
                properties=self._sample_properties(),
                reason=self.rng.choice(["omitted", "silent", "editorial", ""]),
            )

        return Component(
            value=self._sample_value(),
            properties=self._sample_properties(),
        )

    def _sample_value(self) -> Any:
        """Generate a random component value.

        When ``use_pitchspace`` is True, chooses from pitchspace class and
        pair types.  When ``use_generic`` is True, also includes plain ints
        and strings.
        """
        cfg = self.config
        options: list[str] = []
        if cfg.use_pitchspace:
            options.extend(
                [
                    "epc",
                    "spc",
                    "gpc",
                    "eic",
                    "sic",
                    "gic",
                    "ep",
                    "sp",
                    "gp",
                    "ei",
                    "si",
                    "gi",
                ]
            )
        if cfg.use_generic:
            options.extend(["int", "str"])

        if not options:
            return self.rng.randint(0, 127)

        choice = self.rng.choice(options)
        return self._value_from_choice(choice)

    def _value_from_choice(self, choice: str) -> Any:
        """Generate a concrete value from a type choice string."""
        if choice == "int":
            return self.rng.randint(-100, 100)
        if choice == "str":
            labels = ["C", "D", "E", "F", "G", "A", "B", "root", "bass", "x"]
            return self.rng.choice(labels)

        # Pitchspace types — import lazily
        from flexohr.paradigms.pitchspace import (
            EI,
            EIC,
            EP,
            EPC,
            GI,
            GIC,
            GP,
            GPC,
            SI,
            SIC,
            SP,
            SPC,
        )

        # Class types (IntType subclasses)
        if choice == "epc":
            return EPC(self.rng.randint(0, 11))
        if choice == "spc":
            return SPC(self.rng.randint(-7, 7))
        if choice == "gpc":
            return GPC(self.rng.randint(0, 6))
        if choice == "eic":
            return EIC(self.rng.randint(0, 11))
        if choice == "sic":
            return SIC(self.rng.randint(-7, 7))
        if choice == "gic":
            return GIC(self.rng.randint(0, 6))

        # Pair types (PairOfInts subclasses)
        if choice == "ep":
            return EP((self.rng.randint(0, 11), self.rng.randint(0, 8)))
        if choice == "sp":
            return SP((self.rng.randint(-7, 7), self.rng.randint(0, 8)))
        if choice == "gp":
            return GP((self.rng.randint(0, 6), self.rng.randint(0, 8)))
        if choice == "ei":
            return EI((self.rng.randint(0, 11), self.rng.randint(0, 3)))
        if choice == "si":
            return SI((self.rng.randint(-7, 7), self.rng.randint(0, 3)))
        if choice == "gi":
            return GI((self.rng.randint(0, 6), self.rng.randint(0, 3)))

        return self.rng.randint(0, 127)  # fallback

    def _sample_properties(self) -> FrozenProperties:
        """Generate random component-level properties."""
        if self.rng.random() > self.config.property_density:
            return EMPTY_PROPS
        return self._make_random_props()

    def _sample_ohr_properties(self) -> FrozenProperties:
        """Generate random OHR-level properties."""
        if self.rng.random() > self.config.property_density:
            return EMPTY_PROPS

        from flexohr.harmony.harmony_enums import ChordQuality, Inversion

        items: list[Any] = []

        # Optionally add a chord quality
        if self.rng.random() < 0.5:
            qualities = [m for m in ChordQuality if m.name == m.value]
            items.append(self.rng.choice(qualities))

        # Optionally add an inversion
        if self.rng.random() < 0.3:
            inversions = [m for m in Inversion if m.name == m.value]
            items.append(self.rng.choice(inversions))

        if not items:
            return EMPTY_PROPS
        return props(*items)

    def _make_random_props(self) -> FrozenProperties:
        """Build a small random property bag for a component."""
        from flexohr.harmony.enums import ComponentStatus
        from flexohr.harmony.harmony_enums import ToneFunction

        items: list[Any] = []

        # Pick a tone function
        functions = [m for m in ToneFunction if m.name == m.value]
        items.append(self.rng.choice(functions))

        # Optionally add a ComponentStatus
        if self.rng.random() < 0.3:
            statuses = [m for m in ComponentStatus if m.name == m.value]
            items.append(self.rng.choice(statuses))

        return props(*items)

    def _sample_intentionally_invalid(self) -> OHR:
        """Generate an OHR likely to violate common validators.

        Strategies: empty body, missing reference, contradicting tone
        functions on the same component.
        """
        from flexohr.harmony.harmony_enums import ToneFunction

        strategy = self.rng.choice(
            [
                "no_reference",
                "empty_body",
                "contradicting_functions",
            ]
        )

        if strategy == "no_reference":
            # No reference_component at all
            return OHR(
                reference_component=None,
                body=(self._sample_component(),),
            )

        if strategy == "empty_body":
            return OHR(
                reference_component=self._sample_component(),
                body=(),
            )

        # contradicting_functions: put two exclusive ToneFunctions on one component
        exclusive = [
            ToneFunction.root,
            ToneFunction.third,
            ToneFunction.fifth,
            ToneFunction.seventh,
        ]
        tf1, tf2 = self.rng.sample(exclusive, 2)
        bad_comp = Component(
            value=self._sample_value(),
            properties=props(tf1, tf2),
        )
        return OHR(
            reference_component=bad_comp,
            body=(self._sample_component(),),
        )



"""Component types: the atomic building blocks of OHRs.

Paradigm-independent — ``Component.value`` accepts any hashable type.
"""

from __future__ import annotations

from dataclasses import dataclass, replace
from enum import StrEnum
from typing import Any

from flexohr.harmony.enums import (
    EMPTY_PROPS,
    ComponentStatus,
    FrozenProperties,
    KeyValue,
    _property_key,
)


def _coerce_props(raw: Any) -> FrozenProperties:
    """Coerce *raw* to ``FrozenProperties`` (tuple of PropertyItems).

    Accepts:
    - tuple: returned as-is
    - frozenset of (key, value) pairs: converted to tuple of KeyValue/enum items (backward compat)
    - dict of {key: value}: converted to tuple of KeyValue/enum items (backward compat)
    - list/iterable: converted to tuple
    """
    if isinstance(raw, tuple):
        return raw
    if isinstance(raw, frozenset):
        # Backward compat: frozenset of (key, value) pairs -> tuple of items
        result = []
        for item in raw:
            if isinstance(item, tuple) and len(item) == 2:
                key, value = item
                if isinstance(value, StrEnum):
                    result.append(value)
                else:
                    result.append(KeyValue(key=str(key), value=value))
            else:
                result.append(item)
        return tuple(result)
    if isinstance(raw, dict):
        result = []
        for key, value in raw.items():
            if isinstance(value, StrEnum):
                result.append(value)
            else:
                result.append(KeyValue(key=str(key), value=value))
        return tuple(result)
    return tuple(raw)


def _props_repr(properties: FrozenProperties) -> str:
    """Concise text representation of properties, grouping values by key.

    For example, ``(ToneFunction.third, ToneFunction.bass)`` becomes
    ``"tf=third, bass"`` rather than ``"tf=third, tf=bass"``.
    """
    from collections import OrderedDict

    grouped: OrderedDict[str, list[str]] = OrderedDict()
    for item in properties:
        try:
            k = _property_key(item)
        except TypeError:
            k = "?"
        if isinstance(item, KeyValue):
            v_display = repr(item.value)
        elif hasattr(item, "name"):
            v_display = item.name
        else:
            v_display = repr(item)
        # Abbreviate common keys
        abbr = _PROP_KEY_ABBREV.get(k, k)
        grouped.setdefault(abbr, []).append(v_display)

    parts = []
    for k, vals in grouped.items():
        joined = ", ".join(vals)
        parts.append(f"{k}={joined}")
    return ", ".join(parts)


_PROP_KEY_ABBREV: dict[str, str] = {
    "tone_function": "tf",
    "chord_quality": "quality",
    "collection_type": "coll",
    "component_status": "status",
}


def _props_badges_html(
    properties: FrozenProperties,
    bg: str = "#e5e7eb",
    bg_map: dict[str, str] | None = None,
) -> str:
    """Render properties as HTML badges, grouping values under repeated keys.

    For example, ``(ToneFunction.fifth, ToneFunction.bass)`` renders as a
    single badge ``tone_function=fifth, bass`` instead of two separate badges.

    Parameters
    ----------
    properties : FrozenProperties
        The property items to render.
    bg : str
        Default background colour for badges.
    bg_map : dict[str, str] or None
        Optional mapping from property key to background colour.
    """
    from collections import OrderedDict

    if bg_map is None:
        bg_map = {}

    # Group values by key, preserving insertion order
    grouped: OrderedDict[str, list[str]] = OrderedDict()
    for item in properties:
        try:
            k = _property_key(item)
        except TypeError:
            k = "?"
        if isinstance(item, KeyValue):
            v_display = repr(item.value)
        elif hasattr(item, "name"):
            v_display = item.name
        else:
            v_display = repr(item)
        grouped.setdefault(k, []).append(v_display)

    parts = []
    for k, vals in grouped.items():
        colour = bg_map.get(k, bg)
        joined = ", ".join(vals)
        parts.append(
            f'<span style="font-family:monospace;background:{colour};border-radius:3px;'
            f'padding:0 4px;font-size:0.85em;margin-left:2px">{k}={joined}</span>'
        )
    return "".join(parts)


@dataclass(frozen=True, slots=True)
class Component:
    """Immutable wrapper around any value with arbitrary properties.

    Paradigm-independent: *value* can be any hashable type (pitchspace
    Point/Vector, frequency bin, rhythmic pattern, etc.).
    """

    value: Any = None
    properties: FrozenProperties = EMPTY_PROPS
    status: ComponentStatus = ComponentStatus.present

    def __post_init__(self) -> None:
        if not isinstance(self.properties, tuple):
            object.__setattr__(self, "properties", _coerce_props(self.properties))

    def __repr__(self) -> str:
        """Concise representation grouping properties by key.

        Examples::

            Component(SPC(C))
            Component(SIC(M3), tf=third, bass)
            Component(None, tone='root', octave=4)
        """
        parts = [repr(self.value)]
        if self.properties:
            parts.append(", ")
            parts.append(_props_repr(self.properties))
        if self.status is not ComponentStatus.present:
            parts.append(f" [{self.status.value}]")
        return f"C({' '.join(''.join(parts).split())})"

    def __str__(self) -> str:
        return self.__repr__()

    @property
    def is_absent(self) -> bool:
        """True if this is an :class:`AbsentComponent`."""
        return isinstance(self, AbsentComponent)

    @property
    def is_undefined(self) -> bool:
        """True if this is an :class:`UndefinedComponent`."""
        return isinstance(self, UndefinedComponent)

    def get_property(self, key: str, default: Any = None) -> Any:
        """Look up properties by *key*.

        Returns a single value when exactly one match exists, a tuple when
        multiple matches exist, or *default* when no match is found.

        For enum members, the key is the snake_case enum class name
        (e.g., ``"tone_function"`` for ``ToneFunction`` members).
        For KeyValue items, the key is the explicit ``.key`` attribute.
        For dataclass objects, the key is the snake_case class name.
        """
        matches = []
        for item in self.properties:
            try:
                if _property_key(item) == key:
                    if isinstance(item, KeyValue):
                        matches.append(item.value)
                    else:
                        matches.append(item)
            except TypeError:
                continue
        if not matches:
            return default
        if len(matches) == 1:
            return matches[0]
        return tuple(matches)

    def to_dict(self) -> dict[str, Any]:
        """Serialize to a verbose, self-describing dict."""
        from flexohr.core.ohr import _component_to_dict

        return _component_to_dict(self)

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> Component:
        """Deserialize from dict."""
        from flexohr.core.ohr import _component_from_dict

        return _component_from_dict(d)

    def with_(self, **kwargs: Any) -> Component:
        """Return a new Component with specified fields replaced."""
        if "properties" in kwargs:
            p = kwargs["properties"]
            if isinstance(p, dict):
                kwargs["properties"] = _coerce_props(p)
            elif isinstance(p, frozenset):
                kwargs["properties"] = _coerce_props(p)
        return replace(self, **kwargs)

    def with_properties(self, *items: Any, **new_props: Any) -> Component:
        """Return a new Component with additional properties appended.

        Accepts positional items (enum members, dataclass objects) and
        keyword arguments (plain key-value pairs).
        """
        from flexohr.harmony.enums import props as _props

        extra = _props(*items, **new_props)
        return replace(self, properties=self.properties + extra)

    def _repr_html_(self) -> str:
        # Value display
        val = self.value
        if hasattr(val, "_repr_html_"):
            val_html = val._repr_html_()
        else:
            val_html = f'<span style="font-family:monospace">{repr(val)}</span>'

        # Properties badges (grouped by key)
        props_html = _props_badges_html(self.properties, bg="#e5e7eb")

        # Status indicator (only show if not present)
        status_html = ""
        if self.status.value != "present":
            status_html = (
                f'<span style="font-family:monospace;background:#fecaca;border-radius:3px;'
                f'padding:0 4px;font-size:0.85em;margin-left:2px">{self.status.value}</span>'
            )

        return f"{val_html}{props_html}{status_html}"


@dataclass(frozen=True, slots=True)
class AbsentComponent(Component):
    """A component that is structurally expected but explicitly not present.

    Use as ``reference_component`` to signal "no reference" (distinct from
    ``None`` which means "inherit from parent").
    """

    reason: str = ""

    def __post_init__(self) -> None:
        Component.__post_init__(self)
        if self.status is ComponentStatus.present:
            object.__setattr__(self, "status", ComponentStatus.absent)

    def __repr__(self) -> str:
        reason_part = f", reason={self.reason!r}" if self.reason else ""
        return f"AbsentComponent({repr(self.value)}{reason_part})"

    def _repr_html_(self) -> str:
        val = self.value
        if hasattr(val, "_repr_html_"):
            val_html = val._repr_html_()
        else:
            val_html = f'<span style="font-family:monospace">{repr(val)}</span>'

        reason_html = f' <span style="font-style:italic;color:#888">({self.reason})</span>' if self.reason else ""
        return (
            f'<span style="font-family:monospace;color:#dc2626;text-decoration:line-through">'
            f"{val_html}</span>"
            f'<span style="font-family:monospace;background:#fecaca;border-radius:3px;'
            f'padding:0 4px;font-size:0.85em;margin-left:2px">{self.status.value}</span>'
            f"{reason_html}"
        )


@dataclass(frozen=True, slots=True)
class UndefinedComponent(Component):
    """Placeholder component — blocks reference inheritance.

    When used as ``reference_component`` on an OHR, prevents inheriting
    the reference from parent OHRs.  Value is always ``None``.
    """

    def __repr__(self) -> str:
        return "UndefinedComponent()"

    def __post_init__(self) -> None:
        Component.__post_init__(self)
        if self.value is not None:
            object.__setattr__(self, "value", None)


# Module-level sentinel
UNDEFINED = UndefinedComponent()



"""JSON dict list storage backend.

Stores OHRs as a list of nested dicts — the "orientation model" from the
architecture spec.  Queries operate on the dict structures directly.
"""

from __future__ import annotations

import logging
from typing import Any

from flexohr.core.ohr import OHR

logger = logging.getLogger(__name__)


class JsonDictBackend:
    """Store OHRs as a list of nested dicts.

    This is the reference storage model: each OHR is its own
    self-describing dict produced by ``OHR.to_dict()``.
    """

    def __init__(self) -> None:
        self._dicts: list[dict[str, Any]] = []
        self._ohrs: list[OHR] = []

    # ── Ingest ──

    def add_ohr(self, ohr: OHR) -> int:
        """Add an OHR, returning its index."""
        idx = len(self._ohrs)
        self._ohrs.append(ohr)
        self._dicts.append(ohr.to_dict())
        return idx

    def add_ohrs(self, ohrs: list[OHR]) -> list[int]:
        """Add multiple OHRs."""
        return [self.add_ohr(ohr) for ohr in ohrs]

    # ── Access ──

    def get_dict(self, idx: int) -> dict[str, Any]:
        return self._dicts[idx]

    def get_ohr(self, idx: int) -> OHR:
        return self._ohrs[idx]

    def reconstruct_ohr(self, idx: int) -> OHR:
        """Reconstruct OHR from stored dict (tests round-trip)."""
        return OHR.from_dict(self._dicts[idx])

    def reconstruct_all(self) -> list[OHR]:
        """Reconstruct all OHRs from dicts."""
        return [OHR.from_dict(d) for d in self._dicts]

    # ── Queries ──

    def query_ohrs_by_property(self, key: str, value: Any) -> list[int]:
        """Find OHR indices whose top-level properties match."""
        results = []
        for i, d in enumerate(self._dicts):
            for prop in d.get("properties", []):
                if prop.get("type") == "enum" and prop.get("value") == value:
                    results.append(i)
                    break
                if prop.get("type") == "key_value" and prop.get("key") == key and prop.get("value") == value:
                    results.append(i)
                    break
        return results

    def query_components_by_property(self, key: str, value: Any) -> list[dict]:
        """Find all component dicts (across all OHRs) matching a property."""
        results = []
        for d in self._dicts:
            results.extend(self._find_components_in_dict(d, key, value))
        return results

    def query_all_components(self) -> list[dict]:
        """Return all component dicts across all OHRs (flattened)."""
        results = []
        for d in self._dicts:
            results.extend(self._extract_all_components(d))
        return results

    def _extract_all_components(self, d: dict) -> list[dict]:
        """Recursively extract all component dicts from an OHR dict."""
        comps = []
        if d.get("type") in ("component", "absent_component", "undefined_component"):
            comps.append(d)
            return comps
        # It's an OHR dict
        ref = d.get("reference_component")
        if ref is not None:
            comps.append(ref)
        for item in d.get("body", []):
            if item.get("type") == "ohr":
                comps.extend(self._extract_all_components(item))
            else:
                comps.append(item)
        ref_ohr = d.get("reference_ohr")
        if ref_ohr is not None:
            comps.extend(self._extract_all_components(ref_ohr))
        return comps

    def _find_components_in_dict(self, d: dict, key: str, value: Any) -> list[dict]:
        """Find components matching property key/value in an OHR dict tree."""
        results = []
        for comp in self._extract_all_components(d):
            for prop in comp.get("properties", []):
                if prop.get("type") == "enum":
                    # Match by enum class name (snake_cased) or value
                    from flexohr.harmony.enums import _camel_to_snake

                    enum_key = _camel_to_snake(prop.get("enum_class", ""))
                    if enum_key == key and prop.get("value") == value:
                        results.append(comp)
                        break
                elif prop.get("type") == "key_value":
                    if prop.get("key") == key and prop.get("value") == value:
                        results.append(comp)
                        break
        return results

    def count_components(self) -> int:
        """Total component count across all OHRs."""
        return sum(len(self._extract_all_components(d)) for d in self._dicts)

    def count_ohrs(self) -> int:
        """Number of top-level OHRs stored."""
        return len(self._dicts)

    @property
    def memory_bytes(self) -> int:
        """Approximate memory usage."""
        import sys

        return sum(sys.getsizeof(d) for d in self._dicts) + sys.getsizeof(self._ohrs)



"""Central colour registry for ``_repr_html_`` methods.

This module defines the colour scheme for HTML representations of pitchspace types
(pitch, interval, pitch class, interval class) and scale degree types. Colours are
centralized here for consistency and future user customization.

Design Principles
-----------------
1. **Scale degrees** use semantically meaningful colours:

   - **Major** collection types (major, ionian): blue
   - **Minor** collection types (natural_minor, harmonic_minor, etc.): red
   - **Other** collection types (dorian, phrygian, etc.): dark grey
   - **Generic** scale degrees (GSD, GSDI — no collection type): light grey

2. **Pitch and interval types** use three colour families:

   - **Specific** (SPC, SP, SIC, SI): green family
   - **Enharmonic** (EPC, EP, EIC, EI): purple family
   - **Generic** (GPC, GP, GIC, GI): orange family

3. **Within each colour family**, shades distinguish:

   - Pitch class: dark + saturated
   - Pitch (with octave): dark + less saturated
   - Interval class: light + saturated
   - Interval (with octave): light + less saturated

4. **Blue and red are reserved** for scale degrees (major/minor).

Future Customization
--------------------
The dictionaries in this module are designed for future programmatic modification.
Users will be able to override colours at runtime by modifying the registry dicts.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from flexohr.harmony.harmony_enums import CollectionType


# ══════════════════════════════════════════════════════════════════════════════
#  Pitchspace Type Colours
# ══════════════════════════════════════════════════════════════════════════════

#: Colour registry for pitch and interval types.
#: Keys are type abbreviations; values are hex colour strings.
#: Designed for future user customization via direct dict modification.
PITCHSPACE_COLORS: dict[str, str] = {
    # Specific (Green family)
    "SPC": "#166534",  # dark, saturated
    "SP": "#3f6212",  # dark, less saturated
    "SIC": "#4ade80",  # light, saturated
    "SI": "#86efac",  # light, less saturated
    # Enharmonic (Purple family)
    "EPC": "#7c3aed",  # dark, saturated
    "EP": "#6d28d9",  # dark, less saturated
    "EIC": "#c4b5fd",  # light, saturated
    "EI": "#ddd6fe",  # light, less saturated
    # Generic (Orange family)
    "GPC": "#c2410c",  # dark, saturated
    "GP": "#9a3412",  # dark, less saturated
    "GIC": "#fdba74",  # light, saturated
    "GI": "#fed7aa",  # light, less saturated
}


# ══════════════════════════════════════════════════════════════════════════════
#  Scale Degree Colours
# ══════════════════════════════════════════════════════════════════════════════

#: Colour for generic scale degrees (GSD, GSDI) which have no collection type.
GENERIC_SCALE_DEGREE_COLOR: str = "#9ca3af"

#: Colour for minor-mode scale degrees (SD, SDI with minor collection types).
MINOR_SCALE_DEGREE_COLOR: str = "#dc2626"

#: Colour for major-mode scale degrees (SD, SDI with major collection types).
MAJOR_SCALE_DEGREE_COLOR: str = "#2563eb"

#: Default colour for other scale degrees (non-major/minor modes like dorian).
OTHER_SCALE_DEGREE_COLOR: str = "#4b5563"

#: Direct mapping from CollectionType values to scale degree colours.
#: This dict is the primary lookup for ``get_scale_degree_color()``.
#: Designed for future user customization: users can add entries for new
#: collection types or override existing ones.
COLLECTION_TYPE_COLORS: dict[str, str] = {
    # Major family
    "major": MAJOR_SCALE_DEGREE_COLOR,
    "ionian": MAJOR_SCALE_DEGREE_COLOR,
    "M": MAJOR_SCALE_DEGREE_COLOR,
    # Minor family
    "natural_minor": MINOR_SCALE_DEGREE_COLOR,
    "harmonic_minor": MINOR_SCALE_DEGREE_COLOR,
    "melodic_minor": MINOR_SCALE_DEGREE_COLOR,
    "aeolian": MINOR_SCALE_DEGREE_COLOR,
    "m": MINOR_SCALE_DEGREE_COLOR,
    "hm": MINOR_SCALE_DEGREE_COLOR,
    "mm": MINOR_SCALE_DEGREE_COLOR,
    "aeo": MINOR_SCALE_DEGREE_COLOR,
    # Other modes default to OTHER_SCALE_DEGREE_COLOR via fallback
}


def get_scale_degree_color(collection_type: CollectionType | None) -> str:
    """Return the appropriate colour for a scale degree.

    Parameters
    ----------
    collection_type
        The collection type of the scale degree, or ``None`` for generic
        scale degrees (GSD, GSDI).

    Returns
    -------
    str
        Hex colour string (e.g., ``"#2563eb"``).

    Examples
    --------
    >>> from flexohr.harmony.harmony_enums import CollectionType
    >>> get_scale_degree_color(CollectionType.major)
    '#2563eb'
    >>> get_scale_degree_color(CollectionType.natural_minor)
    '#dc2626'
    >>> get_scale_degree_color(CollectionType.dorian)
    '#4b5563'
    >>> get_scale_degree_color(None)
    '#9ca3af'
    """
    if collection_type is None:
        return GENERIC_SCALE_DEGREE_COLOR
    return COLLECTION_TYPE_COLORS.get(collection_type.value, OTHER_SCALE_DEGREE_COLOR)
